from dataclasses import dataclass, field, fields
from enum import IntEnum, StrEnum
from typing import Any, Self

from roborock.data.code_mappings import RoborockProductNickname
from roborock.data.containers import RoborockBase
from roborock.data.v1 import RoborockDockTypeCode


class NewFeatureStrBit(IntEnum):
    TWO_KEY_REAL_TIME_VIDEO = 32
    TWO_KEY_RTV_IN_CHARGING = 33
    DIRTY_REPLENISH_CLEAN = 34
    AUTO_DELIVERY_FIELD_IN_GLOBAL_STATUS = 35
    AVOID_COLLISION_MODE = 36
    VOICE_CONTROL = 37
    NEW_ENDPOINT = 38
    PUMPING_WATER = 39
    CORNER_MOP_STRETCH = 40
    HOT_WASH_TOWEL = 41
    FLOOR_DIR_CLEAN_ANY_TIME = 42
    PET_SUPPLIES_DEEP_CLEAN = 43
    MOP_SHAKE_WATER_MAX = 45
    EXACT_CUSTOM_MODE = 47
    VIDEO_PATROL = 48
    CARPET_CUSTOM_CLEAN = 49
    PET_SNAPSHOT = 50
    CUSTOM_CLEAN_MODE_COUNT = 51
    NEW_AI_RECOGNITION = 52
    AUTO_COLLECTION_2 = 53
    RIGHT_BRUSH_STRETCH = 54
    SMART_CLEAN_MODE_SET = 55
    DIRTY_OBJECT_DETECT = 56
    NO_NEED_CARPET_PRESS_SET = 57
    VOICE_CONTROL_LED = 58
    WATER_LEAK_CHECK = 60
    MIN_BATTERY_15_TO_CLEAN_TASK = 62
    GAP_DEEP_CLEAN = 63
    OBJECT_DETECT_CHECK = 64
    IDENTIFY_ROOM = 66
    MATTER = 67
    WORKDAY_HOLIDAY = 69
    CLEAN_DIRECT_STATUS = 70
    MAP_ERASER = 71
    OPTIMIZE_BATTERY = 72
    ACTIVATE_VIDEO_CHARGING_AND_STANDBY = 73
    CARPET_LONG_HAIRED = 75
    CLEAN_HISTORY_TIME_LINE = 76
    MAX_ZONE_OPENED = 77
    EXHIBITION_FUNCTION = 78
    LDS_LIFTING = 79
    AUTO_TEAR_DOWN_MOP = 80
    SMALL_SIDE_MOP = 81
    SUPPORT_SIDE_BRUSH_UP_DOWN = 82
    DRY_INTERVAL_TIMER = 83
    UVC_STERILIZE = 84
    MIDWAY_BACK_TO_DOCK = 85
    SUPPORT_MAIN_BRUSH_UP_DOWN = 86
    EGG_DANCE_MODE = 87
    MECHANICAL_ARM_MODE = 89
    TIDYUP_ZONES = MECHANICAL_ARM_MODE
    CLEAN_TIME_LINE = 91
    CLEAN_THEN_MOP_MODE = 93
    TYPE_IDENTIFY = 94
    SUPPORT_GET_PARTICULAR_STATUS = 96
    THREE_D_MAPPING_INNER_TEST = 97
    SYNC_SERVER_NAME = 98
    SHOULD_SHOW_ARM_OVER_LOAD = 99
    COLLECT_DUST_COUNT_SHOW = 100
    SUPPORT_API_APP_STOP_GRASP = 101
    CTM_WITH_REPEAT = 102
    SIDE_BRUSH_LIFT_CARPET = 104
    DETECT_WIRE_CARPET = 105
    WATER_SLIDE_MODE = 106
    SOAK_AND_WASH = 107
    CLEAN_EFFICIENCY = 108
    BACK_WASH_NEW_SMART = 109
    DUAL_BAND_WI_FI = 110
    PROGRAM_MODE = 111
    CLEAN_FLUID_DELIVERY = 112
    CARPET_LONG_HAIRED_EX = 113
    OVER_SEA_CTM = 114
    FULL_DUPLES_SWITCH = 115
    LOW_AREA_ACCESS = 116
    FOLLOW_LOW_OBS = 117
    TWO_GEARS_NO_COLLISION = 118
    CARPET_SHAPE_TYPE = 119
    SR_MAP = 120


class ProductFeatures(StrEnum):
    REMOTE_BACK = "remote_back"
    CLEANMODE_MAXPLUS = "cleanmode_maxplus"
    CLEANMODE_PURECLEANMOP = "cleanmode_purecleanmop"
    CLEANMODE_NONE_PURECLEANMOP_WITH_MAXPLUS = "cleanmode_none_purecleanmop_with_maxplus"
    MOP_ELECTRONIC_MODULE = "mop_electronic_module"
    MOP_SHAKE_MODULE = "mop_shake_module"
    MOP_SPIN_MODULE = "mop_spin_module"
    DEFAULT_MAP3D = "map3d"
    DEFAULT_CLEANMODECUSTOM = "custom_cleanmode"
    REALTIMEVIDEO = "realtimevideo"
    REALTIMEVIDEO_LIVECALL = "realtimevideo_livecall"
    REALTIMEVIDEO_RECORDANDSHORTCUT = "realtimevideo_livecall"
    CAMERA_SINGLELINE = "camera_singleline"
    CAMERA_DUALLINE = "camera_dualline"
    CAMERA_RGB = "camera_rgb"
    CAMERA_DOUBLERGB = "camera_doublergb"
    AIRECOGNITION_SETTING = "airecognition_setting"
    AIRECOGNITION_SCENE = "airecognition_scene"
    AIRECOGNITION_PET = "airecognition_pet"
    AIRECOGNITION_OBSTACLE = "airecognition_obstacle"


# The following combinations are pulled directly from decompiled source code.
AIRECOGNITION_OBSTACLE = [ProductFeatures.AIRECOGNITION_OBSTACLE]
RGB_CAMERA_FEATURES = [
    ProductFeatures.CAMERA_RGB,
    ProductFeatures.AIRECOGNITION_SETTING,
    ProductFeatures.AIRECOGNITION_SCENE,
    ProductFeatures.AIRECOGNITION_PET,
    ProductFeatures.AIRECOGNITION_OBSTACLE,
    ProductFeatures.REALTIMEVIDEO,
    ProductFeatures.REALTIMEVIDEO_LIVECALL,
    ProductFeatures.REALTIMEVIDEO_RECORDANDSHORTCUT,
]
DOUBLE_RGB_CAMERA_FEATURES = [
    ProductFeatures.CAMERA_DOUBLERGB,
    ProductFeatures.AIRECOGNITION_SETTING,
    ProductFeatures.AIRECOGNITION_PET,
    ProductFeatures.AIRECOGNITION_OBSTACLE,
    ProductFeatures.REALTIMEVIDEO,
]
SINGLE_LINE_CAMERA_FEATURES = [
    ProductFeatures.CAMERA_SINGLELINE,
    ProductFeatures.AIRECOGNITION_SETTING,
    ProductFeatures.AIRECOGNITION_OBSTACLE,
]
DUAL_LINE_CAMERA_FEATURES = [
    ProductFeatures.CAMERA_DUALLINE,
    ProductFeatures.AIRECOGNITION_SETTING,
    ProductFeatures.AIRECOGNITION_OBSTACLE,
    ProductFeatures.AIRECOGNITION_PET,
]

NEW_DEFAULT_FEATURES = [ProductFeatures.REMOTE_BACK, ProductFeatures.CLEANMODE_MAXPLUS]


PEARL_FEATURES = SINGLE_LINE_CAMERA_FEATURES + [ProductFeatures.CLEANMODE_MAXPLUS, ProductFeatures.MOP_SPIN_MODULE]
PEARL_PLUS_FEATURES = NEW_DEFAULT_FEATURES + RGB_CAMERA_FEATURES + [ProductFeatures.MOP_SPIN_MODULE]
ULTRON_FEATURES = NEW_DEFAULT_FEATURES + DUAL_LINE_CAMERA_FEATURES + [ProductFeatures.MOP_SHAKE_MODULE]
ULTRONSV_FEATURES = NEW_DEFAULT_FEATURES + RGB_CAMERA_FEATURES + [ProductFeatures.MOP_SHAKE_MODULE]
TANOSS_FEATURES = [ProductFeatures.REMOTE_BACK, ProductFeatures.MOP_SHAKE_MODULE]
TOPAZSPOWER_FEATURES = [ProductFeatures.CLEANMODE_MAXPLUS, ProductFeatures.MOP_SHAKE_MODULE]

PRODUCTS_WITHOUT_CUSTOM_CLEAN: set[RoborockProductNickname] = {
    RoborockProductNickname.TANOS,
    RoborockProductNickname.RUBYPLUS,
    RoborockProductNickname.RUBYSC,
    RoborockProductNickname.RUBYSE,
}
PRODUCTS_WITHOUT_DEFAULT_3D_MAP: set[RoborockProductNickname] = {
    RoborockProductNickname.TANOS,
    RoborockProductNickname.TANOSSPLUS,
    RoborockProductNickname.TANOSE,
    RoborockProductNickname.TANOSV,
    RoborockProductNickname.RUBYPLUS,
    RoborockProductNickname.RUBYSC,
    RoborockProductNickname.RUBYSE,
}
PRODUCTS_WITHOUT_PURE_CLEAN_MOP: set[RoborockProductNickname] = {
    RoborockProductNickname.TANOS,
    RoborockProductNickname.TANOSE,
    RoborockProductNickname.TANOSV,
    RoborockProductNickname.TANOSSLITE,
    RoborockProductNickname.TANOSSE,
    RoborockProductNickname.TANOSSC,
    RoborockProductNickname.ULTRONLITE,
    RoborockProductNickname.ULTRONE,
    RoborockProductNickname.RUBYPLUS,
    RoborockProductNickname.RUBYSLITE,
    RoborockProductNickname.RUBYSC,
    RoborockProductNickname.RUBYSE,
}

# Base map containing the initial, unconditional features for each product.
_BASE_PRODUCT_FEATURE_MAP: dict[RoborockProductNickname, list[ProductFeatures]] = {
    RoborockProductNickname.PEARL: PEARL_FEATURES,
    RoborockProductNickname.PEARLS: PEARL_FEATURES,
    RoborockProductNickname.PEARLPLUS: PEARL_PLUS_FEATURES,
    RoborockProductNickname.VIVIAN: PEARL_PLUS_FEATURES,
    RoborockProductNickname.CORAL: PEARL_PLUS_FEATURES,
    RoborockProductNickname.ULTRON: ULTRON_FEATURES,
    RoborockProductNickname.ULTRONE: [ProductFeatures.CLEANMODE_NONE_PURECLEANMOP_WITH_MAXPLUS],
    RoborockProductNickname.ULTRONSV: ULTRONSV_FEATURES,
    RoborockProductNickname.TOPAZSPOWER: TOPAZSPOWER_FEATURES,
    RoborockProductNickname.TANOSS: TANOSS_FEATURES,
    RoborockProductNickname.PEARLC: PEARL_FEATURES,
    RoborockProductNickname.PEARLPLUSS: PEARL_PLUS_FEATURES,
    RoborockProductNickname.PEARLSLITE: PEARL_FEATURES,
    RoborockProductNickname.PEARLE: PEARL_FEATURES,
    RoborockProductNickname.PEARLELITE: PEARL_FEATURES,
    RoborockProductNickname.VIVIANC: [ProductFeatures.CLEANMODE_MAXPLUS, ProductFeatures.MOP_SPIN_MODULE]
    + SINGLE_LINE_CAMERA_FEATURES,
    RoborockProductNickname.CORALPRO: PEARL_PLUS_FEATURES,
    RoborockProductNickname.ULTRONLITE: SINGLE_LINE_CAMERA_FEATURES
    + [ProductFeatures.CLEANMODE_NONE_PURECLEANMOP_WITH_MAXPLUS, ProductFeatures.MOP_ELECTRONIC_MODULE],
    RoborockProductNickname.ULTRONSC: ULTRON_FEATURES,
    RoborockProductNickname.ULTRONSE: [
        ProductFeatures.CLEANMODE_NONE_PURECLEANMOP_WITH_MAXPLUS,
        ProductFeatures.MOP_ELECTRONIC_MODULE,
    ],
    RoborockProductNickname.ULTRONSPLUS: ULTRON_FEATURES,
    RoborockProductNickname.VERDELITE: ULTRONSV_FEATURES,
    RoborockProductNickname.TOPAZS: [ProductFeatures.REMOTE_BACK, ProductFeatures.MOP_SHAKE_MODULE],
    RoborockProductNickname.TOPAZSPLUS: NEW_DEFAULT_FEATURES
    + DUAL_LINE_CAMERA_FEATURES
    + [ProductFeatures.MOP_SHAKE_MODULE],
    RoborockProductNickname.TOPAZSC: TOPAZSPOWER_FEATURES + SINGLE_LINE_CAMERA_FEATURES,
    RoborockProductNickname.TOPAZSV: NEW_DEFAULT_FEATURES + RGB_CAMERA_FEATURES + [ProductFeatures.MOP_SHAKE_MODULE],
    RoborockProductNickname.TANOSSPLUS: TANOSS_FEATURES + DUAL_LINE_CAMERA_FEATURES,
    RoborockProductNickname.TANOSSLITE: [ProductFeatures.MOP_ELECTRONIC_MODULE],
    RoborockProductNickname.TANOSSC: [],
    RoborockProductNickname.TANOSSE: [],
    RoborockProductNickname.TANOSSMAX: NEW_DEFAULT_FEATURES
    + DUAL_LINE_CAMERA_FEATURES
    + [ProductFeatures.MOP_SHAKE_MODULE],
    RoborockProductNickname.TANOS: [ProductFeatures.REMOTE_BACK],
    RoborockProductNickname.TANOSE: [ProductFeatures.MOP_ELECTRONIC_MODULE, ProductFeatures.REMOTE_BACK],
    RoborockProductNickname.TANOSV: DOUBLE_RGB_CAMERA_FEATURES
    + [ProductFeatures.REMOTE_BACK, ProductFeatures.MOP_ELECTRONIC_MODULE],
    RoborockProductNickname.RUBYPLUS: [],
    RoborockProductNickname.RUBYSC: [],
    RoborockProductNickname.RUBYSE: [],
    RoborockProductNickname.RUBYSLITE: [ProductFeatures.MOP_ELECTRONIC_MODULE],
}

PRODUCT_FEATURE_MAP: dict[RoborockProductNickname, list[ProductFeatures]] = {
    product: (
        features
        + ([ProductFeatures.DEFAULT_CLEANMODECUSTOM] if product not in PRODUCTS_WITHOUT_CUSTOM_CLEAN else [])
        + ([ProductFeatures.DEFAULT_MAP3D] if product not in PRODUCTS_WITHOUT_DEFAULT_3D_MAP else [])
        + ([ProductFeatures.CLEANMODE_PURECLEANMOP] if product not in PRODUCTS_WITHOUT_PURE_CLEAN_MOP else [])
    )
    for product, features in _BASE_PRODUCT_FEATURE_MAP.items()
}


@dataclass
class DeviceFeatures(RoborockBase):
    """Represents the features supported by a Roborock device."""

    # Features from robot_new_features (lower 32 bits)
    is_show_clean_finish_reason_supported: bool = field(metadata={"robot_new_features": 1})
    is_re_segment_supported: bool = field(metadata={"robot_new_features": 4})
    is_video_monitor_supported: bool = field(metadata={"robot_new_features": 8})
    is_any_state_transit_goto_supported: bool = field(metadata={"robot_new_features": 16})
    is_fw_filter_obstacle_supported: bool = field(metadata={"robot_new_features": 32})
    is_video_setting_supported: bool = field(metadata={"robot_new_features": 64})
    is_ignore_unknown_map_object_supported: bool = field(metadata={"robot_new_features": 128})
    is_set_child_supported: bool = field(metadata={"robot_new_features": 256})
    is_carpet_supported: bool = field(metadata={"robot_new_features": 512})
    is_record_allowed: bool = field(metadata={"robot_new_features": 1024})
    is_mop_path_supported: bool = field(metadata={"robot_new_features": 2048})
    is_multi_map_segment_timer_supported: bool = field(metadata={"robot_new_features": 4096})
    is_current_map_restore_enabled: bool = field(metadata={"robot_new_features": 8192})
    is_room_name_supported: bool = field(metadata={"robot_new_features": 16384})
    is_shake_mop_set_supported: bool = field(metadata={"robot_new_features": 262144})
    is_map_beautify_internal_debug_supported: bool = field(metadata={"robot_new_features": 2097152})
    is_new_data_for_clean_history: bool = field(metadata={"robot_new_features": 4194304})
    is_new_data_for_clean_history_detail: bool = field(metadata={"robot_new_features": 8388608})
    is_flow_led_setting_supported: bool = field(metadata={"robot_new_features": 16777216})
    is_dust_collection_setting_supported: bool = field(metadata={"robot_new_features": 33554432})
    is_rpc_retry_supported: bool = field(metadata={"robot_new_features": 67108864})
    is_avoid_collision_supported: bool = field(metadata={"robot_new_features": 134217728})
    is_support_set_switch_map_mode: bool = field(metadata={"robot_new_features": 268435456})
    is_map_carpet_add_support: bool = field(metadata={"robot_new_features": 1073741824})
    is_custom_water_box_distance_supported: bool = field(metadata={"robot_new_features": 2147483648})

    # Features from robot_new_features (upper 32 bits)
    is_support_smart_scene: bool = field(metadata={"upper_32_bits": 1})
    is_support_floor_edit: bool = field(metadata={"upper_32_bits": 3})
    is_support_furniture: bool = field(metadata={"upper_32_bits": 4})
    is_wash_then_charge_cmd_supported: bool = field(metadata={"upper_32_bits": 5})
    is_support_room_tag: bool = field(metadata={"upper_32_bits": 6})
    is_support_quick_map_builder: bool = field(metadata={"upper_32_bits": 7})
    is_support_smart_global_clean_with_custom_mode: bool = field(metadata={"upper_32_bits": 8})
    is_careful_slow_mop_supported: bool = field(metadata={"upper_32_bits": 9})
    is_egg_mode_supported_from_new_features: bool = field(metadata={"upper_32_bits": 10})
    is_carpet_show_on_map: bool = field(metadata={"upper_32_bits": 12})
    is_supported_valley_electricity: bool = field(metadata={"upper_32_bits": 13})
    is_unsave_map_reason_supported: bool = field(metadata={"upper_32_bits": 14})
    is_supported_drying: bool = field(metadata={"upper_32_bits": 15})
    is_supported_download_test_voice: bool = field(metadata={"upper_32_bits": 16})
    is_support_backup_map: bool = field(metadata={"upper_32_bits": 17})
    is_support_custom_mode_in_cleaning: bool = field(metadata={"upper_32_bits": 18})
    is_support_remote_control_in_call: bool = field(metadata={"upper_32_bits": 19})

    # Features from new_feature_info_str (masking last 8 chars / 32 bits)
    is_support_set_volume_in_call: bool = field(metadata={"new_feature_str_mask": (1, 8)})
    is_support_clean_estimate: bool = field(metadata={"new_feature_str_mask": (2, 8)})
    is_support_custom_dnd: bool = field(metadata={"new_feature_str_mask": (4, 8)})
    is_carpet_deep_clean_supported: bool = field(metadata={"new_feature_str_mask": (8, 8)})
    is_support_stuck_zone: bool = field(metadata={"new_feature_str_mask": (16, 8)})
    is_support_custom_door_sill: bool = field(metadata={"new_feature_str_mask": (32, 8)})
    is_wifi_manage_supported: bool = field(metadata={"new_feature_str_mask": (128, 8)})
    is_clean_route_fast_mode_supported: bool = field(metadata={"new_feature_str_mask": (256, 8)})
    is_support_cliff_zone: bool = field(metadata={"new_feature_str_mask": (512, 8)})
    is_support_smart_door_sill: bool = field(metadata={"new_feature_str_mask": (1024, 8)})
    is_support_floor_direction: bool = field(metadata={"new_feature_str_mask": (2048, 8)})
    is_back_charge_auto_wash_supported: bool = field(metadata={"new_feature_str_mask": (4096, 8)})
    is_support_incremental_map: bool = field(metadata={"new_feature_str_mask": (4194304, 8)})
    is_offline_map_supported: bool = field(metadata={"new_feature_str_mask": (16384, 8)})
    is_super_deep_wash_supported: bool = field(metadata={"new_feature_str_mask": (32768, 8)})
    is_ces_2022_supported: bool = field(metadata={"new_feature_str_mask": (65536, 8)})
    is_dss_believable: bool = field(metadata={"new_feature_str_mask": (131072, 8)})
    is_main_brush_up_down_supported_from_str: bool = field(metadata={"new_feature_str_mask": (262144, 8)})
    is_goto_pure_clean_path_supported: bool = field(metadata={"new_feature_str_mask": (524288, 8)})
    is_water_up_down_drain_supported: bool = field(metadata={"new_feature_str_mask": (1048576, 8)})
    is_setting_carpet_first_supported: bool = field(metadata={"new_feature_str_mask": (8388608, 8)})
    is_clean_route_deep_slow_plus_supported: bool = field(metadata={"new_feature_str_mask": (16777216, 8)})
    is_dynamically_skip_clean_zone_supported: bool = field(metadata={"new_feature_str_mask": (33554432, 8)})
    is_dynamically_add_clean_zones_supported: bool = field(metadata={"new_feature_str_mask": (67108864, 8)})
    is_left_water_drain_supported: bool = field(metadata={"new_feature_str_mask": (134217728, 8)})
    is_clean_count_setting_supported: bool = field(metadata={"new_feature_str_mask": (1073741824, 8)})
    is_corner_clean_mode_supported: bool = field(metadata={"new_feature_str_mask": (2147483648, 8)})

    # Features from new_feature_info_str (by bit index)
    is_two_key_real_time_video_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.TWO_KEY_REAL_TIME_VIDEO}
    )
    is_two_key_rtv_in_charging_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.TWO_KEY_RTV_IN_CHARGING}
    )
    is_dirty_replenish_clean_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.DIRTY_REPLENISH_CLEAN}
    )
    is_auto_delivery_field_in_global_status_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.AUTO_DELIVERY_FIELD_IN_GLOBAL_STATUS}
    )
    is_avoid_collision_mode_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.AVOID_COLLISION_MODE}
    )
    is_voice_control_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.VOICE_CONTROL})
    is_new_endpoint_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.NEW_ENDPOINT})
    is_pumping_water_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.PUMPING_WATER})
    is_corner_mop_stretch_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.CORNER_MOP_STRETCH})
    is_hot_wash_towel_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.HOT_WASH_TOWEL})
    is_floor_dir_clean_any_time_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.FLOOR_DIR_CLEAN_ANY_TIME}
    )
    is_pet_supplies_deep_clean_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.PET_SUPPLIES_DEEP_CLEAN}
    )
    is_mop_shake_water_max_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.MOP_SHAKE_WATER_MAX}
    )
    is_exact_custom_mode_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.EXACT_CUSTOM_MODE})
    is_video_patrol_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.VIDEO_PATROL})
    is_carpet_custom_clean_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CARPET_CUSTOM_CLEAN}
    )
    is_pet_snapshot_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.PET_SNAPSHOT})
    is_custom_clean_mode_count_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CUSTOM_CLEAN_MODE_COUNT}
    )
    is_new_ai_recognition_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.NEW_AI_RECOGNITION})
    is_auto_collection_2_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.AUTO_COLLECTION_2})
    is_right_brush_stretch_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.RIGHT_BRUSH_STRETCH}
    )
    is_smart_clean_mode_set_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SMART_CLEAN_MODE_SET}
    )
    is_dirty_object_detect_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.DIRTY_OBJECT_DETECT}
    )
    is_no_need_carpet_press_set_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.NO_NEED_CARPET_PRESS_SET}
    )
    is_voice_control_led_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.VOICE_CONTROL_LED})
    is_water_leak_check_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.WATER_LEAK_CHECK})
    is_min_battery_15_to_clean_task_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.MIN_BATTERY_15_TO_CLEAN_TASK}
    )
    is_gap_deep_clean_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.GAP_DEEP_CLEAN})
    is_object_detect_check_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.OBJECT_DETECT_CHECK}
    )
    is_identify_room_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.IDENTIFY_ROOM})
    is_matter_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.MATTER})
    is_workday_holiday_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.WORKDAY_HOLIDAY})
    is_clean_direct_status_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CLEAN_DIRECT_STATUS}
    )
    is_map_eraser_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.MAP_ERASER})
    is_optimize_battery_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.OPTIMIZE_BATTERY})
    is_activate_video_charging_and_standby_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.ACTIVATE_VIDEO_CHARGING_AND_STANDBY}
    )
    is_carpet_long_haired_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.CARPET_LONG_HAIRED})
    is_clean_history_time_line_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CLEAN_HISTORY_TIME_LINE}
    )
    is_max_zone_opened_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.MAX_ZONE_OPENED})
    is_exhibition_function_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.EXHIBITION_FUNCTION}
    )
    is_lds_lifting_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.LDS_LIFTING})
    is_auto_tear_down_mop_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.AUTO_TEAR_DOWN_MOP})
    is_small_side_mop_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.SMALL_SIDE_MOP})
    is_support_side_brush_up_down_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SUPPORT_SIDE_BRUSH_UP_DOWN}
    )
    is_dry_interval_timer_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.DRY_INTERVAL_TIMER})
    is_uvc_sterilize_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.UVC_STERILIZE})
    is_midway_back_to_dock_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.MIDWAY_BACK_TO_DOCK}
    )
    is_support_main_brush_up_down_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SUPPORT_MAIN_BRUSH_UP_DOWN}
    )
    is_egg_dance_mode_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.EGG_DANCE_MODE})
    is_mechanical_arm_mode_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.MECHANICAL_ARM_MODE}
    )
    is_tidyup_zones_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.TIDYUP_ZONES})
    is_clean_time_line_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.CLEAN_TIME_LINE})
    is_clean_then_mop_mode_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CLEAN_THEN_MOP_MODE}
    )
    is_type_identify_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.TYPE_IDENTIFY})
    is_support_get_particular_status_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SUPPORT_GET_PARTICULAR_STATUS}
    )
    is_three_d_mapping_inner_test_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.THREE_D_MAPPING_INNER_TEST}
    )
    is_sync_server_name_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.SYNC_SERVER_NAME})
    is_should_show_arm_over_load_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SHOULD_SHOW_ARM_OVER_LOAD}
    )
    is_collect_dust_count_show_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.COLLECT_DUST_COUNT_SHOW}
    )
    is_support_api_app_stop_grasp_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SUPPORT_API_APP_STOP_GRASP}
    )
    is_ctm_with_repeat_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.CTM_WITH_REPEAT})
    is_side_brush_lift_carpet_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.SIDE_BRUSH_LIFT_CARPET}
    )
    is_detect_wire_carpet_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.DETECT_WIRE_CARPET})
    is_water_slide_mode_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.WATER_SLIDE_MODE})
    is_soak_and_wash_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.SOAK_AND_WASH})
    is_clean_efficiency_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.CLEAN_EFFICIENCY})
    is_back_wash_new_smart_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.BACK_WASH_NEW_SMART}
    )
    is_dual_band_wi_fi_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.DUAL_BAND_WI_FI})
    is_program_mode_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.PROGRAM_MODE})
    is_clean_fluid_delivery_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CLEAN_FLUID_DELIVERY}
    )
    is_carpet_long_haired_ex_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.CARPET_LONG_HAIRED_EX}
    )
    is_over_sea_ctm_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.OVER_SEA_CTM})
    is_full_duples_switch_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.FULL_DUPLES_SWITCH})
    is_low_area_access_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.LOW_AREA_ACCESS})
    is_follow_low_obs_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.FOLLOW_LOW_OBS})
    is_two_gears_no_collision_supported: bool = field(
        metadata={"new_feature_str_bit": NewFeatureStrBit.TWO_GEARS_NO_COLLISION}
    )
    is_carpet_shape_type_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.CARPET_SHAPE_TYPE})
    is_sr_map_supported: bool = field(metadata={"new_feature_str_bit": NewFeatureStrBit.SR_MAP})

    # Features from feature_info list
    is_led_status_switch_supported: bool = field(metadata={"robot_features": 119})
    is_multi_floor_supported: bool = field(metadata={"robot_features": 120})
    is_support_fetch_timer_summary: bool = field(metadata={"robot_features": 122})
    is_order_clean_supported: bool = field(metadata={"robot_features": 123})
    is_analysis_supported: bool = field(metadata={"robot_features": 124})
    is_remote_supported: bool = field(metadata={"robot_features": 125})
    is_support_voice_control_debug: bool = field(metadata={"robot_features": 130})

    # Features from model whitelists/blacklists or other flags
    is_mop_forbidden_supported: bool = field(
        metadata={
            "model_whitelist": [
                RoborockProductNickname.TANOSV,
                RoborockProductNickname.TOPAZSV,
                RoborockProductNickname.TANOS,
                RoborockProductNickname.TANOSE,
                RoborockProductNickname.TANOSSLITE,
                RoborockProductNickname.TANOSS,
                RoborockProductNickname.TANOSSPLUS,
                RoborockProductNickname.TANOSSMAX,
                RoborockProductNickname.ULTRON,
                RoborockProductNickname.ULTRONLITE,
                RoborockProductNickname.PEARL,
                RoborockProductNickname.RUBYSLITE,
            ]
        }
    )
    is_soft_clean_mode_supported: bool = field(
        metadata={
            "model_whitelist": [
                RoborockProductNickname.TANOSV,
                RoborockProductNickname.TANOSE,
                RoborockProductNickname.TANOS,
            ]
        }
    )
    is_custom_mode_supported: bool = field(metadata={"model_blacklist": [RoborockProductNickname.TANOS]})
    is_support_custom_carpet: bool = field(metadata={"model_whitelist": [RoborockProductNickname.ULTRONLITE]})
    is_show_general_obstacle_supported: bool = field(metadata={"model_whitelist": [RoborockProductNickname.TANOSSPLUS]})
    is_show_obstacle_photo_supported: bool = field(
        metadata={
            "model_whitelist": [
                RoborockProductNickname.TANOSSPLUS,
                RoborockProductNickname.TANOSSMAX,
                RoborockProductNickname.ULTRON,
            ]
        }
    )
    is_rubber_brush_carpet_supported: bool = field(metadata={"model_whitelist": [RoborockProductNickname.ULTRONLITE]})
    is_carpet_pressure_use_origin_paras_supported: bool = field(
        metadata={"model_whitelist": [RoborockProductNickname.ULTRONLITE]}
    )
    is_support_mop_back_pwm_set: bool = field(metadata={"model_whitelist": [RoborockProductNickname.PEARL]})
    is_collect_dust_mode_supported: bool = field(metadata={"model_blacklist": [RoborockProductNickname.PEARL]})
    is_support_water_mode: bool = field(
        metadata={
            "product_features": [
                ProductFeatures.MOP_ELECTRONIC_MODULE,
                ProductFeatures.MOP_SHAKE_MODULE,
                ProductFeatures.MOP_SPIN_MODULE,
            ]
        }
    )
    is_pure_clean_mop_supported: bool = field(metadata={"product_features": [ProductFeatures.CLEANMODE_PURECLEANMOP]})
    is_new_remote_view_supported: bool = field(metadata={"product_features": [ProductFeatures.REMOTE_BACK]})
    is_max_plus_mode_supported: bool = field(metadata={"product_features": [ProductFeatures.CLEANMODE_MAXPLUS]})
    is_none_pure_clean_mop_with_max_plus: bool = field(
        metadata={"product_features": [ProductFeatures.CLEANMODE_NONE_PURECLEANMOP_WITH_MAXPLUS]}
    )
    is_clean_route_setting_supported: bool = field(
        metadata={"product_features": [ProductFeatures.MOP_SHAKE_MODULE, ProductFeatures.MOP_SPIN_MODULE]}
    )
    is_mop_shake_module_supported: bool = field(metadata={"product_features": [ProductFeatures.MOP_SHAKE_MODULE]})
    is_customized_clean_supported: bool = field(
        metadata={"product_features": [ProductFeatures.MOP_SHAKE_MODULE, ProductFeatures.MOP_SPIN_MODULE]}
    )

    # Raw feature info values from get_init_status for diagnostics
    new_feature_info: int = field(default=0, repr=False)
    new_feature_info_str: str = field(default="", repr=False)
    feature_info: list[int] = field(default_factory=list, repr=False)

    @classmethod
    def from_feature_flags(
        cls,
        new_feature_info: int,
        new_feature_info_str: str,
        feature_info: list[int],
        product_nickname: RoborockProductNickname | None,
    ) -> Self:
        """Creates a DeviceFeatures instance from raw feature flags.
        :param new_feature_info: A int from get_init_status (sometimes can be found in homedata, but it is not always)
        :param new_feature_info_str: A hex string from get_init_status or home_data.
        :param feature_info: A list of ints from get_init_status
        :param product_nickname: The product nickname of the device."""
        # For any future reverse engineerining:
        # RobotNewFeatures = new_feature_info
        # newFeatureInfoStr = new_feature_info_str
        # feature_info =robotFeatures
        kwargs: dict[str, Any] = {
            # Store raw feature info for diagnostics
            "new_feature_info": new_feature_info,
            "new_feature_info_str": new_feature_info_str,
            "feature_info": feature_info,
        }

        for f in fields(cls):
            # Skip raw feature info fields (already set above)
            if f.name in ("new_feature_info", "new_feature_info_str", "feature_info"):
                continue
            # Default all features to False.
            kwargs[f.name] = False
            if not f.metadata:
                continue

            if (mask := f.metadata.get("robot_new_features")) is not None:
                kwargs[f.name] = bool(mask & new_feature_info)
            elif (bit_index := f.metadata.get("upper_32_bits")) is not None:
                # Check bits in the upper 32-bit integer of new_feature_info
                if new_feature_info:
                    kwargs[f.name] = bool(((new_feature_info >> 32) >> bit_index) & 1)
            elif (mask_info := f.metadata.get("new_feature_str_mask")) is not None:
                # Check bitmask against a slice of the hex string
                if new_feature_info_str:
                    try:
                        mask, slice_count = mask_info
                        if len(new_feature_info_str) >= slice_count:
                            last_chars = new_feature_info_str[-slice_count:]
                            value = int(last_chars, 16)
                            kwargs[f.name] = bool(mask & value)
                    except (ValueError, IndexError):
                        pass  # Keep it False
            elif (bit := f.metadata.get("new_feature_str_bit")) is not None:
                # Check a specific bit in the hex string using its index
                if new_feature_info_str:
                    try:
                        # Bit index defines which character and which bit inside it to check
                        char_index_from_end = 1 + bit.value // 4
                        if char_index_from_end <= len(new_feature_info_str):
                            char_hex = new_feature_info_str[-char_index_from_end]
                            nibble = int(char_hex, 16)
                            bit_in_nibble = bit.value % 4
                            kwargs[f.name] = bool((nibble >> bit_in_nibble) & 1)
                    except (ValueError, IndexError):
                        pass  # Keep it False
            elif (feature_id := f.metadata.get("robot_features")) is not None:
                kwargs[f.name] = feature_id in feature_info
            elif (whitelist := f.metadata.get("model_whitelist")) is not None:
                # If product_nickname is None, assume it is not in the whitelist
                kwargs[f.name] = product_nickname in whitelist or product_nickname is None
            elif (blacklist := f.metadata.get("model_blacklist")) is not None:
                # If product_nickname is None, assume it is not in the blacklist.
                if product_nickname is None:
                    kwargs[f.name] = True
                else:
                    kwargs[f.name] = product_nickname not in blacklist
            elif (product_features := f.metadata.get("product_features")) is not None:
                if product_nickname is not None:
                    available_features = PRODUCT_FEATURE_MAP.get(product_nickname, [])
                    if any(feat in available_features for feat in product_features):  # type: ignore
                        kwargs[f.name] = True

        return cls(**kwargs)

    def get_supported_features(self) -> list[str]:
        """Returns a list of supported features (Primarily used for logging purposes)."""
        return [k for k, v in vars(self).items() if v]


WASH_N_FILL_DOCK_TYPES = [
    RoborockDockTypeCode.empty_wash_fill_dock,
    RoborockDockTypeCode.s8_dock,
    RoborockDockTypeCode.p10_dock,
    RoborockDockTypeCode.p10_pro_dock,
    RoborockDockTypeCode.s8_maxv_ultra_dock,
    RoborockDockTypeCode.qrevo_s_dock,
    RoborockDockTypeCode.saros_r10_dock,
    RoborockDockTypeCode.qrevo_curv_dock,
    RoborockDockTypeCode.qrevo_s5v_dock,
    RoborockDockTypeCode.saros_20_dock,
]


def is_wash_n_fill_dock(dock_type: RoborockDockTypeCode) -> bool:
    """Check if the dock type is a wash and fill dock."""
    return dock_type in WASH_N_FILL_DOCK_TYPES


def is_valid_dock(dock_type: RoborockDockTypeCode) -> bool:
    """Check if device supports a dock."""
    return dock_type != RoborockDockTypeCode.no_dock



"""Common test utils for the protocols package."""

import json
from typing import Any

from Crypto.Cipher import AES
from Crypto.Util.Padding import pad

from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol


def build_a01_message(message: dict[Any, Any], seq: int = 2020) -> RoborockMessage:
    """Build an encoded A01 RPC response message."""
    return RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_RESPONSE,
        payload=pad(
            json.dumps(
                {
                    "dps": message,  # {10000: json.dumps(message)},
                }
            ).encode(),
            AES.block_size,
        ),
        version=b"A01",
        seq=seq,
    )



from roborock.protocol import Utils


def test_encryption():
    """Tests the L01 GCM encryption logic."""
    local_key = "b8Hj5mFk3QzT7rLp"
    timestamp = 1753606905
    sequence = 1
    nonce = 304251
    connect_nonce = 893563
    ack_nonce = 485592656
    payload_str = (
        '{"dps":{"101":"{\\"id\\":1806,\\"method\\":\\"get_prop\\",\\"params\\":[\\"get_status\\"]}"},"t":1753606905}'
    )
    payload = payload_str.encode("utf-8")

    encrypted_data = Utils.encrypt_gcm_l01(
        plaintext=payload,
        local_key=local_key,
        timestamp=timestamp,
        sequence=sequence,
        nonce=nonce,
        connect_nonce=connect_nonce,
        ack_nonce=ack_nonce,
    )

    expected_data = bytes.fromhex(
        "fd60c8daca1ccae67f6077477bfa9d37189a38d75b3c4a907c2435d3c146ee84d8f99597e3e1571a015961ceaa4d64bc3695fae024c341"
        "6737d77150341de29cad2f95bfaf532358f12bbff89f140fef5b1ee284c3abfe3b83a577910a72056dab4d5a75b182d1a0cba145e3e450"
        "f3927443"
    )

    assert encrypted_data == expected_data


def test_decryption():
    """Tests the L01 GCM decryption logic."""
    local_key = "b8Hj5mFk3QzT7rLp"
    timestamp = 1753606905
    sequence = 1
    nonce = 304251
    connect_nonce = 893563
    ack_nonce = 485592656
    payload = bytes.fromhex(
        "fd60c8daca1ccae67f6077477bfa9d37189a38d75b3c4a907c2435d3c146ee84d8f99597e3e1571a015961ceaa4d64bc3695fae024c341"
        "6737d77150341de29cad2f95bfaf532358f12bbff89f140fef5b1ee284c3abfe3b83a577910a72056dab4d5a75b182d1a0cba145e3e450"
        "f3927443"
    )
    decrypted_data = Utils.decrypt_gcm_l01(
        payload=payload,
        local_key=local_key,
        timestamp=timestamp,
        sequence=sequence,
        nonce=nonce,
        connect_nonce=connect_nonce,
        ack_nonce=ack_nonce,
    )
    decrypted_str = decrypted_data.decode("utf-8")

    expected_str = (
        '{"dps":{"101":"{\\"id\\":1806,\\"method\\":\\"get_prop\\",\\"params\\":[\\"get_status\\"]}"},"t":1753606905}'
    )
    assert decrypted_str == expected_str



"""Tests for the DustCollectionModeTrait class."""

from unittest.mock import AsyncMock, call

import pytest

from roborock.data import RoborockDockDustCollectionModeCode, RoborockDockTypeCode
from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.dust_collection_mode import DustCollectionModeTrait
from roborock.roborock_typing import RoborockCommand

DUST_COLLECTION_MODE_DATA = [{"mode": 2}]


@pytest.fixture(name="dust_collection_mode")
def dust_collection_mode_trait(
    device: RoborockDevice,
    discover_features_fixture: None,
) -> DustCollectionModeTrait | None:
    """Create a DustCollectionModeTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.dust_collection_mode


@pytest.mark.parametrize(
    ("dock_type_code"),
    [
        (RoborockDockTypeCode.s7_max_ultra_dock),
        (RoborockDockTypeCode.s8_dock),
        (RoborockDockTypeCode.p10_dock),
        (RoborockDockTypeCode.qrevo_s_dock),
        (RoborockDockTypeCode.qrevo_s5v_dock),
        (RoborockDockTypeCode.saros_20_dock),
    ],
)
async def test_dust_collection_mode_available(
    dust_collection_mode: DustCollectionModeTrait | None,
    mock_rpc_channel: AsyncMock,
    dock_type_code: RoborockDockTypeCode,
) -> None:
    """Test successfully refreshing the dust collection mode."""
    assert dust_collection_mode is not None

    mock_rpc_channel.send_command.side_effect = [
        DUST_COLLECTION_MODE_DATA,
    ]

    await dust_collection_mode.refresh()

    mock_rpc_channel.send_command.assert_has_calls(
        [
            call(RoborockCommand.GET_DUST_COLLECTION_MODE),
        ]
    )

    assert dust_collection_mode.mode == RoborockDockDustCollectionModeCode.balanced


@pytest.mark.parametrize(("dock_type_code"), [(RoborockDockTypeCode.no_dock)])
async def test_unsupported_dust_collection_mode(
    dust_collection_mode: DustCollectionModeTrait | None,
    dock_type_code: RoborockDockTypeCode,
) -> None:
    """Test that the trait is not available for unsupported dock types."""
    assert dust_collection_mode is None



"""Testing fakes and simulators for python-roborock.

This package provides stateful firmware simulators (e.g. `V1VacuumSimulator`),
fake transport channels (`FakeChannel`), and cloud orchestration simulators (`FakeRoborockCloud`)
to allow downstream consumers (such as Home Assistant integrations) to write high-fidelity
integration tests using the real client library classes instead of fragile top-level mocks.

Testing Architecture & Boundaries
---------------------------------
We fake communication at two boundaries:
1.  **Network HTTP API Interception**: `FakeRoborockCloud.patch_device_manager()` routes
    HTTP requests (such as discovery, login, home details) to custom mock endpoints using
    `aioresponses` under the hood. No Python client methods are mocked; the real EAPI client
    executes fully.
2.  **Plaintext RPC Message Interception**: Device communication is intercepted at the
    plaintext JSON RPC level (Layer 2). The real client classes (`V1Channel`, `MqttChannel`)
    run under test, but their transport calls are intercepted by our stateful simulators.

     ┌────────────────────────────────────────────────────────┐
     │               TESTED CLIENT (REAL CODE)                │
     │                                                        │
     │  RoborockDevice / Traits / V1RpcChannel / V1Channel    │
     └──────────────────────────┬─────────────────────────────┘
                                │
                      ROBOROCKMESSAGE PAYLOADS
                      (Plaintext JSON commands)
                                │
     ┌──────────────────────────▼─────────────────────────────┐
     │                 SIMULATOR (TEST FAKE)                  │
     │                                                        │
     │  FakeChannel (Intercepts publish/subscribe)            │
     │  RoborockDeviceSimulator (Stateful firmware simulator) │
     └────────────────────────────────────────────────────────┘

Integration Usage Example
-------------------------
```python
from roborock.testing import FakeRoborockCloud, V1VacuumSimulator

async def test_start_vacuum_service():
    # Setup cloud state and add a simulated vacuum device
    cloud = FakeRoborockCloud()
    fake_device = V1VacuumSimulator(duid="living_room_s7", battery=100, state=RoborockStateCode.charging)
    cloud.add_device(fake_device)

    # Patch channels and API calls using our cloud context manager
    with cloud.patch_device_manager():
        # Create the real client manager (logins and discovers natively via mock HTTP)
        manager = await create_device_manager(
            user_params=UserParams(username="test_user", user_data=USER_DATA),
            cache=InMemoryCache(),
        )

    # Fetch the discovered device client
    devices = await manager.get_devices()
    device = devices[0]

    # Trigger client start command
    await device.v1_properties.command.send("app_start")

    # Assert against the simulated vacuum state
    assert fake_device.state == RoborockStateCode.cleaning
```
"""

from roborock.testing.channel import FakeChannel
from roborock.testing.cloud import FakeRoborockCloud, FakeWebApiClient
from roborock.testing.simulator import (
    DEFAULT_KEY_T,
    DEFAULT_LOCAL_KEY,
    DEFAULT_PRODUCT_ID,
    RoborockDeviceSimulator,
)
from roborock.testing.v1_simulator import (
    DEFAULT_APP_INIT,
    DEFAULT_CLEAN_SUMMARY,
    DEFAULT_CONSUMABLE,
    DEFAULT_DND_TIMER,
    DEFAULT_LAST_CLEAN_RECORD,
    DEFAULT_NETWORK_INFO,
    DEFAULT_STATUS,
    V1VacuumSimulator,
)

__all__ = [
    "DEFAULT_APP_INIT",
    "DEFAULT_CLEAN_SUMMARY",
    "DEFAULT_CONSUMABLE",
    "DEFAULT_DND_TIMER",
    "DEFAULT_KEY_T",
    "DEFAULT_LAST_CLEAN_RECORD",
    "DEFAULT_LOCAL_KEY",
    "DEFAULT_NETWORK_INFO",
    "DEFAULT_PRODUCT_ID",
    "DEFAULT_STATUS",
    "FakeChannel",
    "FakeRoborockCloud",
    "FakeWebApiClient",
    "RoborockDeviceSimulator",
    "V1VacuumSimulator",
]



"""End-to-end tests for LocalChannel using fake sockets."""

import asyncio
from collections.abc import AsyncGenerator

import pytest
import syrupy

from roborock.devices.transport.local_channel import LocalChannel
from roborock.protocol import MessageParser, create_local_decoder
from roborock.protocols.v1_protocol import LocalProtocolVersion
from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol
from tests.fixtures.logging import CapturedRequestLog
from tests.fixtures.mqtt import Subscriber
from tests.mock_data import LOCAL_KEY

TEST_HOST = "192.168.1.100"
TEST_DEVICE_UID = "test_device_uid"
TEST_RANDOM = 23


@pytest.fixture(name="local_channel")
async def local_channel_fixture(mock_async_create_local_connection: None) -> AsyncGenerator[LocalChannel, None]:
    channel = LocalChannel(host=TEST_HOST, local_key=LOCAL_KEY, device_uid=TEST_DEVICE_UID)
    yield channel
    channel.close()


def build_raw_response(
    protocol: RoborockMessageProtocol,
    seq: int,
    payload: bytes,
    version: LocalProtocolVersion = LocalProtocolVersion.V1,
    connect_nonce: int | None = None,
    ack_nonce: int | None = None,
) -> bytes:
    """Build an encoded response message."""
    message = RoborockMessage(
        protocol=protocol,
        random=23,
        seq=seq,
        payload=payload,
        version=version.value.encode(),
    )
    return MessageParser.build(message, local_key=LOCAL_KEY, connect_nonce=connect_nonce, ack_nonce=ack_nonce)


async def test_connect(
    local_channel: LocalChannel,
    local_response_queue: asyncio.Queue[bytes],
    local_received_requests: asyncio.Queue[bytes],
    log: CapturedRequestLog,
    snapshot: syrupy.SnapshotAssertion,
) -> None:
    """Test connecting to the device."""
    # Queue HELLO response with payload to ensure it can be parsed
    local_response_queue.put_nowait(build_raw_response(RoborockMessageProtocol.HELLO_RESPONSE, 1, payload=b"ok"))

    await local_channel.connect()

    assert local_channel.is_connected
    assert local_received_requests.qsize() == 1

    # Verify HELLO request
    request_bytes = await local_received_requests.get()
    # Note: We cannot use create_local_decoder here because HELLO_REQUEST has payload=None
    # which causes MessageParser to fail parsing. For now we verify the raw bytes.

    # Protocol is at offset 19 (2 bytes)
    # Prefix(4) + Version(3) + Seq(4) + Random(4) + Timestamp(4) = 19
    assert len(request_bytes) >= 21
    protocol_bytes = request_bytes[19:21]
    assert int.from_bytes(protocol_bytes, "big") == RoborockMessageProtocol.HELLO_REQUEST

    assert snapshot == log


async def test_send_command(
    local_channel: LocalChannel,
    local_response_queue: asyncio.Queue[bytes],
    local_received_requests: asyncio.Queue[bytes],
    log: CapturedRequestLog,
    snapshot: syrupy.SnapshotAssertion,
) -> None:
    """Test sending a command."""
    # Queue HELLO response
    local_response_queue.put_nowait(build_raw_response(RoborockMessageProtocol.HELLO_RESPONSE, 1, payload=b"ok"))

    await local_channel.connect()

    # Clear requests from handshake
    while not local_received_requests.empty():
        await local_received_requests.get()

    # Send command
    cmd_seq = 123
    msg = RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_REQUEST,
        seq=cmd_seq,
        payload=b'{"method":"get_status"}',
    )
    # Prepare a fake response to the command.
    local_response_queue.put_nowait(
        build_raw_response(RoborockMessageProtocol.RPC_RESPONSE, cmd_seq, payload=b'{"status": "ok"}')
    )

    subscriber = Subscriber()
    unsub = await local_channel.subscribe(subscriber.append)

    await local_channel.publish(msg)

    # Verify request received by the server
    request_bytes = await local_received_requests.get()
    assert local_received_requests.empty()

    # Decode request
    decoder = create_local_decoder(local_key=LOCAL_KEY)
    msgs = list(decoder(request_bytes))
    assert len(msgs) == 1
    assert msgs[0].protocol == RoborockMessageProtocol.RPC_REQUEST
    assert msgs[0].payload == b'{"method":"get_status"}'
    assert msgs[0].version == LocalProtocolVersion.V1.value.encode()

    # Verify response received by subscriber
    await subscriber.wait()
    assert len(subscriber.messages) == 1
    response_message = subscriber.messages[0]
    assert isinstance(response_message, RoborockMessage)
    assert response_message.protocol == RoborockMessageProtocol.RPC_RESPONSE
    assert response_message.payload == b'{"status": "ok"}'

    unsub()

    assert snapshot == log


async def test_l01_session(
    local_channel: LocalChannel,
    local_response_queue: asyncio.Queue[bytes],
    local_received_requests: asyncio.Queue[bytes],
    log: CapturedRequestLog,
    snapshot: syrupy.SnapshotAssertion,
) -> None:
    """Test connecting to a device that speaks the L01 protocol.

    Note that this test currently has a delay because the actual local client
    will delay before retrying with L01 after a failed 1.0 attempt. This should
    also be improved in the actual client itself, but likely requires a closer
    look at the actual device response in that scenario or moving to a serial
    request/response behavior rather than publish/subscribe.
    """
    # Client first attempts 1.0 and we reply with a fake invalid response. The
    # response is arbitrary, and this could be improved by capturing a real L01
    # device response to a 1.0 message.
    local_response_queue.put_nowait(b"\x00")
    # The client attempts L01 protocol as a followup. The connect nonce uses
    # a deterministic number from deterministic_message_fixtures.
    connect_nonce = 9090
    local_response_queue.put_nowait(
        build_raw_response(
            RoborockMessageProtocol.HELLO_RESPONSE,
            1,
            payload=b"ok",
            version=LocalProtocolVersion.L01,
            connect_nonce=connect_nonce,
            ack_nonce=None,
        )
    )

    await local_channel.connect()

    assert local_channel.is_connected

    # Verify 1.0 HELLO request
    request_bytes = await local_received_requests.get()
    # Protocol is at offset 19 (2 bytes)
    # Prefix(4) + Version(3) + Seq(4) + Random(4) + Timestamp(4) = 19
    assert len(request_bytes) >= 21
    protocol_bytes = request_bytes[19:21]
    assert int.from_bytes(protocol_bytes, "big") == RoborockMessageProtocol.HELLO_REQUEST

    # Verify L01 HELLO request
    request_bytes = await local_received_requests.get()
    # Protocol is at offset 19 (2 bytes)
    # Prefix(4) + Version(3) + Seq(4) + Random(4) + Timestamp(4) = 19
    assert len(request_bytes) >= 21
    protocol_bytes = request_bytes[19:21]
    assert int.from_bytes(protocol_bytes, "big") == RoborockMessageProtocol.HELLO_REQUEST

    assert local_received_requests.empty()

    # Verify the channel switched to L01 protocol
    assert local_channel.protocol_version == LocalProtocolVersion.L01.value

    # We have established a connection. Now send some messages.
    # Publish an L01 command. Currently the caller of the local channel needs to
    # determine the protocol version to use, but this could be pushed inside of
    # the channel in the future.
    cmd_seq = 123
    msg = RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_REQUEST,
        seq=cmd_seq,
        payload=b'{"method":"get_status"}',
        version=b"L01",
    )
    # Prepare a fake response to the command.
    local_response_queue.put_nowait(
        build_raw_response(
            RoborockMessageProtocol.RPC_RESPONSE,
            cmd_seq,
            payload=b'{"status": "ok"}',
            version=LocalProtocolVersion.L01,
            connect_nonce=connect_nonce,
            ack_nonce=TEST_RANDOM,
        )
    )

    # Set up a subscriber to listen for the response then publish the message.
    subscriber = Subscriber()
    unsub = await local_channel.subscribe(subscriber.append)
    await local_channel.publish(msg)

    # Verify request received by the server
    request_bytes = await local_received_requests.get()
    decoder = create_local_decoder(local_key=LOCAL_KEY, connect_nonce=connect_nonce, ack_nonce=TEST_RANDOM)
    msgs = list(decoder(request_bytes))
    assert len(msgs) == 1
    assert msgs[0].protocol == RoborockMessageProtocol.RPC_REQUEST
    assert msgs[0].payload == b'{"method":"get_status"}'
    assert msgs[0].version == LocalProtocolVersion.L01.value.encode()

    # Verify fake response published by the server, received by subscriber
    await subscriber.wait()
    assert len(subscriber.messages) == 1
    response_message = subscriber.messages[0]
    assert isinstance(response_message, RoborockMessage)
    assert response_message.protocol == RoborockMessageProtocol.RPC_RESPONSE
    assert response_message.payload == b'{"status": "ok"}'
    assert response_message.version == LocalProtocolVersion.L01.value.encode()

    unsub()

    assert snapshot == log



"""Tests for the MapContentTrait."""

from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.map_content import MapContentTrait
from roborock.map.map_parser import ParsedMapData
from roborock.roborock_typing import RoborockCommand


@pytest.fixture
def map_content_trait(device: RoborockDevice) -> MapContentTrait:
    """Create a MapContentTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.map_content


async def test_refresh_map_content_trait(
    map_content_trait: MapContentTrait,
    mock_map_rpc_channel: AsyncMock,
) -> None:
    """Test successfully getting and parsing map content."""
    map_data = b"dummy_map_bytes"
    mock_map_rpc_channel.send_command.return_value = map_data
    mock_parsed_data = ParsedMapData(
        image_content=b"dummy_image_content",
        map_data=MagicMock(),
    )

    with patch("roborock.devices.traits.v1.map_content.MapParser.parse", return_value=mock_parsed_data) as mock_parse:
        await map_content_trait.refresh()
        mock_parse.assert_called_once_with(map_data)

    assert map_content_trait.image_content == b"dummy_image_content"
    assert map_content_trait.map_data is not None

    mock_map_rpc_channel.send_command.assert_called_once_with(RoborockCommand.GET_MAP_V1)



"""Tests for the B01 protocol message encoding and decoding."""

import json
import logging
import pathlib
from collections.abc import Generator
from typing import Any

import pytest
from freezegun import freeze_time
from syrupy import SnapshotAssertion

from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP, YXWaterLevel
from roborock.data.code_mappings import completed_warnings
from roborock.exceptions import RoborockException
from roborock.map.b01_q10_map_parser import Q10MapPacket, Q10TracePacket
from roborock.protocols.b01_q10_protocol import (
    Q10DpsUpdate,
    decode_message,
    decode_rpc_response,
    encode_mqtt_payload,
)
from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol

TESTDATA_PATH = pathlib.Path("tests/protocols/testdata/b01_q10_protocol/")
TESTDATA_FILES = list(TESTDATA_PATH.glob("*.json"))
TESTDATA_IDS = [x.stem for x in TESTDATA_FILES]

MAP_FIXTURE = pathlib.Path("tests/map/testdata/b01_q10_map.bin")
TRACE_FIXTURE = pathlib.Path("tests/map/testdata/b01_q10_trace.bin")


def _message(payload: bytes, protocol: RoborockMessageProtocol) -> RoborockMessage:
    return RoborockMessage(protocol=protocol, payload=payload, version=b"B01")


def test_decode_message_dps_update() -> None:
    """A non-MAP_RESPONSE message decodes into a Q10DpsUpdate."""
    message = _message(b'{"dps": {"122": 100}}', RoborockMessageProtocol.RPC_RESPONSE)
    decoded = decode_message(message)
    assert decoded == Q10DpsUpdate(dps={B01_Q10_DP.BATTERY: 100})


def test_decode_message_map_packet() -> None:
    """A MAP_RESPONSE 01 01 payload decodes into a Q10MapPacket."""
    message = _message(MAP_FIXTURE.read_bytes(), RoborockMessageProtocol.MAP_RESPONSE)
    decoded = decode_message(message)
    assert isinstance(decoded, Q10MapPacket)
    assert {room.id: room.name for room in decoded.rooms} == {2: "Living Room", 3: "Bedroom"}


def test_decode_message_trace_packet() -> None:
    """A MAP_RESPONSE 02 01 payload decodes into a Q10TracePacket."""
    message = _message(TRACE_FIXTURE.read_bytes(), RoborockMessageProtocol.MAP_RESPONSE)
    decoded = decode_message(message)
    assert isinstance(decoded, Q10TracePacket)
    assert [(p.x, p.y) for p in decoded.points] == [(169, 0)]


def test_decode_message_unknown_map_marker_returns_none() -> None:
    """A MAP_RESPONSE with an unrecognized marker is skipped (returns None)."""
    assert decode_message(_message(b"\x09\x09junk", RoborockMessageProtocol.MAP_RESPONSE)) is None
    assert decode_message(_message(b"", RoborockMessageProtocol.MAP_RESPONSE)) is None


@pytest.fixture(autouse=True)
def fixed_time_fixture() -> Generator[None, None, None]:
    """Fixture to freeze time for predictable request IDs."""
    with freeze_time("2025-01-20T12:00:00"):
        yield


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_decode_rpc_payload(filename: str, snapshot: SnapshotAssertion) -> None:
    """Test decoding a B01 RPC response protocol message."""
    with open(filename, "rb") as f:
        payload = f.read()

    message = RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"B01",
        random=97431,
        timestamp=1652547161,
    )

    decoded_message = decode_rpc_response(message)
    assert json.dumps(decoded_message, indent=2) == snapshot


@pytest.mark.parametrize(
    ("payload", "expected_error_message"),
    [
        (b"", "missing payload"),
        (b"n", "Invalid B01 json payload"),
        (b"{}", "missing 'dps'"),
        (b'{"dps": []}', "'dps' should be a dictionary"),
        (b'{"dps": {"not_a_number": 123}}', "dps key is not a valid integer"),
        (b'{"dps": {"101": 123}}', "Invalid dpCommon format: expected dict"),
        (b'{"dps": {"101": {"not_a_number": 123}}}', "Invalid dpCommon format: dps key is not a valid intege"),
    ],
)
def test_decode_invalid_rpc_payload(payload: bytes, expected_error_message: str) -> None:
    """Test decoding a B01 RPC response protocol message."""
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"B01",
        random=97431,
        timestamp=1652547161,
    )
    with pytest.raises(RoborockException, match=expected_error_message):
        decode_rpc_response(message)


def test_decode_unknown_dps_code(caplog: pytest.LogCaptureFixture) -> None:
    """Unknown data points are dropped silently, without logging warnings.

    ss07 hardware pushes DPs 112 and 113 (and occasionally others) that this
    library does not model. They must be ignored without emitting "not a valid
    code" warnings, which previously spammed the log on every status push.
    """
    completed_warnings.discard("112 is not a valid code for B01_Q10_DP")
    completed_warnings.discard("113 is not a valid code for B01_Q10_DP")
    completed_warnings.discard("909090 is not a valid code for B01_Q10_DP")
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.RPC_RESPONSE,
        payload=b'{"dps": {"909090": 123, "112": 0, "113": 0, "122": 100}}',
        seq=12750,
        version=b"B01",
        random=97431,
        timestamp=1652547161,
    )

    with caplog.at_level(logging.WARNING):
        decoded_message = decode_rpc_response(message)
    assert decoded_message == {
        B01_Q10_DP.BATTERY: 100,
    }
    assert "not a valid code" not in caplog.text


@pytest.mark.parametrize(
    ("command", "params"),
    [
        (B01_Q10_DP.REQUEST_DPS, {}),
        (B01_Q10_DP.REQUEST_DPS, None),
        (B01_Q10_DP.START_CLEAN, {"cmd": 1}),
        (B01_Q10_DP.WATER_LEVEL, YXWaterLevel.MEDIUM.code),
    ],
)
def test_encode_mqtt_payload(command: B01_Q10_DP, params: dict[str, Any], snapshot) -> None:
    """Test encoding of MQTT payload for B01 Q10 commands."""

    message = encode_mqtt_payload(command, params)
    assert isinstance(message, RoborockMessage)
    assert message.protocol == RoborockMessageProtocol.RPC_REQUEST
    assert message.version == b"B01"
    assert message.payload is not None

    # Snapshot the raw payload to ensure stable encoding. We verify it is
    # valid json
    assert snapshot == message.payload

    json.loads(message.payload.decode())



"""Consumable / accessory life trait for Q10 B01 devices."""

import logging

from roborock.data.b01_q10.b01_q10_containers import Q10Consumable
from roborock.devices.traits.common import DpsDataConverter

from .common import UpdatableTrait

_LOGGER = logging.getLogger(__name__)


class ConsumableTrait(Q10Consumable, UpdatableTrait):
    """Trait exposing remaining life of consumables (brushes, filter, sensors)."""

    _CONVERTER = DpsDataConverter.from_dataclass(Q10Consumable)

    def __init__(self) -> None:
        """Initialize the consumable trait."""
        Q10Consumable.__init__(self)
        UpdatableTrait.__init__(self, command=None, logger=_LOGGER)



"""Module for Roborock devices.

This interface is experimental and subject to breaking changes without notice
until the API is stable.
"""

import asyncio
import datetime
import logging
from abc import ABC
from collections.abc import Callable
from typing import Any

from roborock.callbacks import CallbackList
from roborock.data import HomeDataDevice, HomeDataProduct
from roborock.diagnostics import redact_device_data
from roborock.exceptions import RoborockException
from roborock.roborock_message import RoborockMessage
from roborock.util import RoborockLoggerAdapter

from .traits import Trait
from .traits.traits_mixin import TraitsMixin
from .transport.channel import Channel

_LOGGER = logging.getLogger(__name__)

__all__ = [
    "DeviceReadyCallback",
    "RoborockDevice",
]

# Exponential backoff parameters
MIN_BACKOFF_INTERVAL = datetime.timedelta(seconds=10)
MAX_BACKOFF_INTERVAL = datetime.timedelta(minutes=30)
BACKOFF_MULTIPLIER = 1.5
# Give time for the NETWORK_INFO fetch and V1 hello attempt
# and potential fallback to L01.
START_ATTEMPT_TIMEOUT = datetime.timedelta(seconds=15)


DeviceReadyCallback = Callable[["RoborockDevice"], None]


class RoborockDevice(ABC, TraitsMixin):
    """A generic channel for establishing a connection with a Roborock device.

    Individual channel implementations have their own methods for speaking to
    the device that hide some of the protocol specific complexity, but they
    are still specialized for the device type and protocol.

    Attributes of the device are exposed through traits, which are mixed in
    through the TraitsMixin class. Traits are optional and may not be present
    on all devices.
    """

    def __init__(
        self,
        device_info: HomeDataDevice,
        product: HomeDataProduct,
        channel: Channel,
        trait: Trait,
    ) -> None:
        """Initialize the RoborockDevice.

        The device takes ownership of the channel for communication with the device.
        Use `connect()` to establish the connection, which will set up the appropriate
        protocol channel. Use `close()` to clean up all connections.
        """
        TraitsMixin.__init__(self, trait)
        self._duid = device_info.duid
        self._logger = RoborockLoggerAdapter(duid=self._duid, logger=_LOGGER)
        self._name = device_info.name
        self._device_info = device_info
        self._product = product
        self._channel = channel
        self._connect_task: asyncio.Task[None] | None = None
        self._unsub: Callable[[], None] | None = None
        self._ready_callbacks = CallbackList["RoborockDevice"]()
        self._has_connected = False

    @property
    def duid(self) -> str:
        """Return the device unique identifier (DUID)."""
        return self._duid

    @property
    def name(self) -> str:
        """Return the device name."""
        return self._name

    @property
    def device_info(self) -> HomeDataDevice:
        """Return the device information.

        This includes information specific to the device like its identifier or
        firmware version.
        """
        return self._device_info

    @property
    def product(self) -> HomeDataProduct:
        """Return the device product name.

        This returns product level information such as the model name.
        """
        return self._product

    @property
    def is_connected(self) -> bool:
        """Return whether the device is connected."""
        return self._channel.is_connected

    @property
    def is_local_connected(self) -> bool:
        """Return whether the device is connected locally.

        This can be used to determine if the device is reachable over a local
        network connection, as opposed to a cloud connection. This is useful
        for adjusting behavior like polling frequency.
        """
        return self._channel.is_local_connected

    def add_ready_callback(self, callback: DeviceReadyCallback) -> Callable[[], None]:
        """Add a callback to be notified when the device is ready.

        A device is considered ready when it has successfully connected. It may go
        offline later, but this callback will only be called once when the device
        first connects.

        The callback will be called immediately if the device has already previously
        connected.
        """
        remove = self._ready_callbacks.add_callback(callback)
        if self._has_connected:
            callback(self)

        return remove

    async def start_connect(self) -> None:
        """Start a background task to connect to the device.

        This will give a moment for the first connection attempt to start so
        that the device will have connections established -- however, this will
        never directly fail.

        If the connection fails, it will retry in the background with
        exponential backoff.

        Once connected, the device will remain connected until `close()` is
        called. The device will automatically attempt to reconnect if the connection
        is lost.
        """
        # The future will be set to True if the first attempt succeeds, False if
        # it fails, or an exception if an unexpected error occurs.
        # We use this to wait a short time for the first attempt to complete. We
        # don't actually care about the result, just that we waited long enough.
        start_attempt: asyncio.Future[bool] = asyncio.Future()

        async def connect_loop() -> None:
            try:
                backoff = MIN_BACKOFF_INTERVAL
                while True:
                    try:
                        await self.connect()
                        if not start_attempt.done():
                            start_attempt.set_result(True)
                        self._has_connected = True
                        self._ready_callbacks(self)
                        return
                    except RoborockException as e:
                        if not start_attempt.done():
                            start_attempt.set_result(False)
                        self._logger.info("Failed to connect (retry %s): %s", backoff.total_seconds(), e)
                        await asyncio.sleep(backoff.total_seconds())
                        backoff = min(backoff * BACKOFF_MULTIPLIER, MAX_BACKOFF_INTERVAL)
                    except Exception as e:  # pylint: disable=broad-except
                        if not start_attempt.done():
                            start_attempt.set_exception(e)
                        self._logger.exception("Uncaught error during connect: %s", e)
                        return
            except asyncio.CancelledError:
                self._logger.debug("connect_loop was cancelled for device %s", self.duid)
            finally:
                if not start_attempt.done():
                    start_attempt.set_result(False)

        self._connect_task = asyncio.create_task(connect_loop())

        try:
            async with asyncio.timeout(START_ATTEMPT_TIMEOUT.total_seconds()):
                await start_attempt
        except TimeoutError:
            self._logger.debug("Initial connection attempt took longer than expected, will keep trying in background")

    async def connect(self) -> None:
        """Connect to the device using the appropriate protocol channel."""
        if self._unsub:
            raise ValueError("Already connected to the device")
        unsub = await self._channel.subscribe(self._on_message)
        try:
            if self.v1_properties is not None:
                await self.v1_properties.start()
            elif self.b01_q10_properties is not None:
                await self.b01_q10_properties.start()
        except RoborockException:
            # Expected: start() can fail transiently. Unsubscribe before propagating
            # so the retry by connect_loop() gets a clean channel.
            unsub()
            raise
        except Exception:
            # Not expected here. We normally avoid a bare ``except Exception`` in
            # this codebase, but a leaked subscription would stop connect_loop() from
            # ever reconnecting, so we deliberately catch broadly, log the unexpected
            # error, and release the channel before propagating.
            self._logger.exception("Unexpected error during connect; releasing channel to avoid a leak")
            unsub()
            raise
        self._logger.info("Connected to device")
        self._unsub = unsub

    async def close(self) -> None:
        """Close all connections to the device."""
        if self._connect_task:
            self._connect_task.cancel()
            try:
                await self._connect_task
            except asyncio.CancelledError:
                pass
        if self.v1_properties is not None:
            self.v1_properties.close()
        if self.b01_q10_properties is not None:
            await self.b01_q10_properties.close()
        if self._unsub:
            self._unsub()
            self._unsub = None

    def _on_message(self, message: RoborockMessage) -> None:
        """Handle incoming messages from the device."""
        self._logger.debug("Received message from device: %s", message)

    def diagnostic_data(self) -> dict[str, Any]:
        """Return diagnostics information about the device."""
        extra: dict[str, Any] = {}
        if self.v1_properties:
            extra["traits"] = self.v1_properties.as_dict()
        return redact_device_data(
            {
                "device": self.device_info.as_dict(),
                "product": self.product.as_dict(),
                **extra,
            }
        )



"""Tests for the v1 protocol message encoding and decoding."""

import json
import logging
import pathlib
from collections.abc import Generator
from unittest.mock import patch

import pytest
from freezegun import freeze_time
from syrupy import SnapshotAssertion

from roborock.data.containers import RoborockBase, UserData
from roborock.exceptions import RoborockException
from roborock.protocol import Utils
from roborock.protocols.v1_protocol import (
    RequestMessage,
    SecurityData,
    create_map_response_decoder,
    decode_data_protocol_message,
    decode_rpc_response,
)
from roborock.roborock_message import RoborockDataProtocol, RoborockMessage, RoborockMessageProtocol
from roborock.roborock_typing import RoborockCommand
from tests import mock_data

USER_DATA = UserData.from_dict(mock_data.USER_DATA)
TEST_REQUEST_ID = 44444
TEST_ENDPOINT = "87ItGWdb"
TEST_ENDPOINT_BYTES = TEST_ENDPOINT.encode()
SECURITY_DATA = SecurityData(
    endpoint=TEST_ENDPOINT,
    nonce=b"\x91\xbe\x10\xc9b+\x9d\x8a\xcdH*\x19\xf6\xfe\x81h",
)


TESTDATA_PATH = pathlib.Path("tests/protocols/testdata/v1_protocol/")
TESTDATA_FILES = list(TESTDATA_PATH.glob("*.json"))
TESTDATA_IDS = [x.stem for x in TESTDATA_FILES]


@pytest.fixture(autouse=True)
def fixed_time_fixture() -> Generator[None, None, None]:
    """Fixture to freeze time for predictable request IDs."""
    # Freeze time to a specific point so request IDs are predictable
    with freeze_time("2025-01-20T12:00:00"):
        yield


@pytest.fixture(name="test_request_id", autouse=True)
def request_id_fixture() -> Generator[int, None, None]:
    """Fixture to provide a fixed request ID."""
    with patch("roborock.protocols.v1_protocol.get_next_int", return_value=TEST_REQUEST_ID):
        yield TEST_REQUEST_ID


@pytest.mark.parametrize(
    ("command", "params", "expected"),
    [
        (
            RoborockCommand.GET_STATUS,
            None,
            b'{"dps":{"101":"{\\"id\\":44444,\\"method\\":\\"get_status\\",\\"params\\":[]}"},"t":1737374400}',
        )
    ],
)
def test_encode_local_payload(command, params, expected):
    """Test encoding of local payload for V1 commands."""
    message = RequestMessage(command, params).encode_message(RoborockMessageProtocol.GENERAL_REQUEST)
    assert isinstance(message, RoborockMessage)
    assert message.protocol == RoborockMessageProtocol.GENERAL_REQUEST
    assert message.payload == expected


@pytest.mark.parametrize(
    ("command", "params", "expected"),
    [
        (
            RoborockCommand.GET_STATUS,
            None,
            b'{"dps":{"101":"{\\"id\\":44444,\\"method\\":\\"get_status\\",\\"params\\":[],\\"security\\":{\\"endpoint\\":\\"87ItGWdb\\",\\"nonce\\":\\"91be10c9622b9d8acd482a19f6fe8168\\"}}"},"t":1737374400}',
        )
    ],
)
def test_encode_mqtt_payload(command, params, expected):
    """Test encoding of local payload for V1 commands."""
    request_message = RequestMessage(command, params=params)
    message = request_message.encode_message(RoborockMessageProtocol.RPC_REQUEST, SECURITY_DATA)
    assert isinstance(message, RoborockMessage)
    assert message.protocol == RoborockMessageProtocol.RPC_REQUEST
    assert message.payload == expected


@pytest.mark.parametrize(
    ("payload", "expected"),
    [
        (
            b'{"t":1652547161,"dps":{"102":"{\\"id\\":20005,\\"result\\":[{\\"msg_ver\\":2,\\"msg_seq\\":1072,\\"state\\":8,\\"battery\\":100,\\"clean_time\\":1041,\\"clean_area\\":37080000,\\"error_code\\":0,\\"map_present\\":1,\\"in_cleaning\\":0,\\"in_returning\\":0,\\"in_fresh_state\\":1,\\"lab_status\\":1,\\"water_box_status\\":0,\\"fan_power\\":103,\\"dnd_enabled\\":0,\\"map_status\\":3,\\"is_locating\\":0,\\"lock_status\\":0,\\"water_box_mode\\":202,\\"distance_off\\":0,\\"water_box_carriage_status\\":0,\\"mop_forbidden_enable\\":0,\\"unsave_map_reason\\":0,\\"unsave_map_flag\\":0}]}"}}',
            [
                {
                    "msg_ver": 2,
                    "msg_seq": 1072,
                    "state": 8,
                    "battery": 100,
                    "clean_time": 1041,
                    "clean_area": 37080000,
                    "error_code": 0,
                    "map_present": 1,
                    "in_cleaning": 0,
                    "in_returning": 0,
                    "in_fresh_state": 1,
                    "lab_status": 1,
                    "water_box_status": 0,
                    "fan_power": 103,
                    "dnd_enabled": 0,
                    "map_status": 3,
                    "is_locating": 0,
                    "lock_status": 0,
                    "water_box_mode": 202,
                    "distance_off": 0,
                    "water_box_carriage_status": 0,
                    "mop_forbidden_enable": 0,
                    "unsave_map_reason": 0,
                    "unsave_map_flag": 0,
                }
            ],
        ),
    ],
)
def test_decode_rpc_response(payload: bytes, expected: RoborockBase) -> None:
    """Test decoding a v1 RPC response protocol message."""
    # The values other than the payload are arbitrary
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )
    decoded_message = decode_rpc_response(message)
    assert decoded_message.request_id == 20005
    assert decoded_message.data == expected


@pytest.mark.parametrize("filename", TESTDATA_FILES, ids=TESTDATA_IDS)
def test_decode_rpc_payload(filename: str, snapshot: SnapshotAssertion) -> None:
    """Test decoding a v1 RPC response protocol message."""
    with open(filename, "rb") as f:
        payload = f.read()
    # The values other than the payload are arbitrary
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )
    decoded_message = decode_rpc_response(message)
    assert decoded_message.request_id == snapshot
    assert json.dumps(decoded_message.data, indent=2) == snapshot


def test_create_map_response_decoder():
    """Test creating and using a map response decoder."""
    test_data = b"some map\n"
    compressed_data = (
        b"\x1f\x8b\x08\x08\xf9\x13\x99h\x00\x03foo\x00+\xce\xcfMU\xc8M,\xe0\x02\x00@\xdb\xc6\x1a\t\x00\x00\x00"
    )

    # Create header: endpoint(8) + padding(8) + request_id(2) + padding(6)
    # request_id = 44508 (0xaddc in little endian)
    header = TEST_ENDPOINT_BYTES + b"\x00" * 8 + b"\xdc\xad" + b"\x00" * 6
    encrypted_data = Utils.encrypt_cbc(compressed_data, SECURITY_DATA.nonce)
    payload = header + encrypted_data

    message = RoborockMessage(
        protocol=RoborockMessageProtocol.MAP_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )

    decoder = create_map_response_decoder(SECURITY_DATA)
    result = decoder(message)
    assert result is not None
    assert result.request_id == 44508
    assert result.data == test_data


def test_create_map_response_decoder_invalid_endpoint(caplog: pytest.LogCaptureFixture):
    """Test map response decoder with invalid endpoint."""
    caplog.set_level(logging.DEBUG)
    # Create header with wrong endpoint
    header = b"wrongend" + b"\x00" * 8 + b"\xdc\xad" + b"\x00" * 6
    payload = header + b"encrypted_data"

    message = RoborockMessage(
        protocol=RoborockMessageProtocol.MAP_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )

    decoder = create_map_response_decoder(SECURITY_DATA)
    assert decoder(message) is None
    assert "Received map response not requested by this device, ignoring." in caplog.text


def test_create_map_response_decoder_invalid_payload():
    """Test map response decoder with invalid payload."""
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.MAP_RESPONSE,
        payload=b"short",  # Too short payload
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )

    decoder = create_map_response_decoder(SECURITY_DATA)

    with pytest.raises(RoborockException, match="Invalid V1 map response format: missing payload"):
        decoder(message)


@pytest.mark.parametrize(
    ("payload", "expected_data", "expected_error"),
    [
        (
            b'{"t":1757883536,"dps":{"102":"{\\"id\\":20001,\\"result\\":\\"unknown_method\\"}"}}',
            {},
            "The method called is not recognized by the device.",
        ),
        (
            b'{"t":1757883536,"dps":{"102":"{\\"id\\":20001,\\"result\\":\\"other\\"}"}}',
            {},
            "Unexpected API Result",
        ),
    ],
)
def test_decode_result_with_error(payload: bytes, expected_data: dict[str, str], expected_error: str) -> None:
    """Test decoding a v1 RPC response protocol message."""
    # The values other than the payload are arbitrary
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )
    decoded_message = decode_rpc_response(message)
    assert decoded_message.request_id == 20001
    assert decoded_message.data == expected_data
    assert decoded_message.api_error
    assert expected_error in str(decoded_message.api_error)


def test_decode_no_request_id():
    """Test map response decoder without a request id is raised as an exception."""
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=b'{"t":1757883536,"dps":{"102":"{\\"result\\":\\"unknown_method\\"}"}}',
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )
    with pytest.raises(RoborockException, match="The method called is not recognized by the device"):
        decode_rpc_response(message)


def test_decode_error_without_result() -> None:
    """Test decoding an error-only V1 RPC response (no 'result' key)."""
    payload = (
        b'{"t":1768419740,"dps":{"102":"{\\"id\\":20062,\\"error\\":{\\"code\\":-10007,'
        b'\\"message\\":\\"invalid status\\"}}"}}'
    )
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1768419740,
    )
    decoded_message = decode_rpc_response(message)
    assert decoded_message.request_id == 20062
    assert decoded_message.data == {}
    assert decoded_message.api_error
    assert isinstance(decoded_message.api_error.args[0], dict)
    assert decoded_message.api_error.args[0]["code"] == -10007


def test_invalid_unicode() -> None:
    """Test an error while decoding unicode bytes"""
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=b"hello\xff\xfe",  # Invalid UTF-8 bytes
        seq=12750,
        version=b"1.0",
        random=97431,
        timestamp=1652547161,
    )
    with pytest.raises(RoborockException, match="Invalid V1 message payload"):
        decode_rpc_response(message)


def test_decode_data_protocol_message() -> None:
    """Test decoding a V1 push message with data protocol updates."""
    payload = json.dumps({"t": 1652547161, "dps": {"121": 8, "122": 95}}).encode()
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
    )
    result = decode_data_protocol_message(message)
    assert result == {RoborockDataProtocol.STATE: 8, RoborockDataProtocol.BATTERY: 95}


def test_decode_data_protocol_message_all_status_fields() -> None:
    """Test decoding a push message with all known status data protocol fields."""
    payload = json.dumps(
        {
            "t": 1652547161,
            "dps": {"120": 0, "121": 5, "122": 100, "123": 102, "124": 204, "133": 1},
        }
    ).encode()
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
    )
    result = decode_data_protocol_message(message)
    assert result == {
        RoborockDataProtocol.ERROR_CODE: 0,
        RoborockDataProtocol.STATE: 5,
        RoborockDataProtocol.BATTERY: 100,
        RoborockDataProtocol.FAN_POWER: 102,
        RoborockDataProtocol.WATER_BOX_MODE: 204,
        RoborockDataProtocol.CHARGE_STATUS: 1,
    }


def test_decode_data_protocol_message_unknown_codes() -> None:
    """Test that unknown data protocol codes are ignored."""
    payload = json.dumps({"t": 1652547161, "dps": {"121": 8, "999": 42}}).encode()
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
    )
    result = decode_data_protocol_message(message)
    assert result == {RoborockDataProtocol.STATE: 8}


def test_decode_data_protocol_message_empty_payload() -> None:
    """Test decoding with empty payload returns None."""
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=None,
    )
    assert decode_data_protocol_message(message) is None


def test_decode_data_protocol_message_rpc_response() -> None:
    """Test that an RPC response (code 102) produces None since the value is not a data protocol."""
    # This contains an RPC response (102) which has a JSON string value, not a data protocol code
    payload = json.dumps(
        {
            "t": 1652547161,
            "dps": {"102": '{"id":20001,"result":[{"state":8}]}'},
        }
    ).encode()
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
    )
    # Code 102 is not in RoborockDataProtocol enum, so it should be ignored.
    # The result should be None (no recognized data protocol codes).
    assert decode_data_protocol_message(message) is None


def test_decode_data_protocol_message_no_dps() -> None:
    """Test decoding message without dps returns None."""
    payload = json.dumps({"t": 1652547161}).encode()
    message = RoborockMessage(
        protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
        payload=payload,
    )
    assert decode_data_protocol_message(message) is None



"""Tests for the map parser."""

from pathlib import Path

import pytest

from roborock.exceptions import RoborockException
from roborock.map.map_parser import MapParser, MapParserConfig

MAP_DATA_FILE = Path(__file__).parent / "raw_map_data"
DEFAULT_MAP_CONFIG = MapParserConfig()


@pytest.mark.parametrize("map_content", [b"", b"12345"])
def test_invalid_map_content(map_content: bytes):
    """Test that parsing map data returns the expected image and data."""
    parser = MapParser(DEFAULT_MAP_CONFIG)
    with pytest.raises(RoborockException, match="Failed to parse map data"):
        parser.parse(map_content)


# We can add additional tests here in the future that actually parse valid map data



from dataclasses import Field, fields

from roborock.data import AppInitStatus, HomeDataProduct, RoborockBase
from roborock.data.v1.v1_containers import FieldNameBase
from roborock.device_features import DeviceFeatures
from roborock.devices.cache import DeviceCache
from roborock.devices.traits.v1 import common
from roborock.roborock_typing import RoborockCommand


class DeviceTraitsConverter(common.V1TraitDataConverter):
    """Converter for APP_GET_INIT_STATUS responses into DeviceFeatures."""

    def __init__(self, product: HomeDataProduct) -> None:
        """Initialize DeviceTraitsConverter."""
        self._product = product

    def convert(self, response: common.V1ResponseData) -> DeviceFeatures:
        """Parse an APP_GET_INIT_STATUS response into a DeviceFeatures instance."""
        if not isinstance(response, list):
            raise ValueError(f"Unexpected AppInitStatus response format: {type(response)}: {response!r}")
        app_status = AppInitStatus.from_dict(response[0])
        return DeviceFeatures.from_feature_flags(
            new_feature_info=app_status.new_feature_info,
            new_feature_info_str=app_status.new_feature_info_str,
            feature_info=app_status.feature_info,
            product_nickname=self._product.product_nickname,
        )


class DeviceFeaturesTrait(DeviceFeatures, common.V1TraitMixin):
    """Trait for managing supported features on Roborock devices."""

    command = RoborockCommand.APP_GET_INIT_STATUS
    converter: DeviceTraitsConverter

    def __init__(self, product: HomeDataProduct, device_cache: DeviceCache) -> None:  # pylint: disable=super-init-not-called
        """Initialize DeviceFeaturesTrait."""
        common.V1TraitMixin.__init__(self)
        self.converter = DeviceTraitsConverter(product)
        self._product = product
        self._device_cache = device_cache
        # All fields of DeviceFeatures are required. Initialize them to False
        # so we have some known state.
        for field in fields(self):
            setattr(self, field.name, False)

    @staticmethod
    def _get_dataclass_field(cls: type[RoborockBase], field_name: FieldNameBase) -> Field:
        """Look up a dataclass field by its FieldNameBase name."""
        for f in fields(cls):
            if f.name == field_name:
                return f
        raise ValueError(f"Field {field_name!r} not found in {cls}")

    def is_field_supported(self, cls: type[RoborockBase], field_name: FieldNameBase) -> bool:
        """Determines if the specified field is supported by this device.

        We use the `dps` dataclass field metadata to get the `RoborockDataProtocol`
        integer ID and check it against the set of supported schema IDs for the
        device returned in the product information.
        """
        dataclass_field = self._get_dataclass_field(cls, field_name)
        if (dps := dataclass_field.metadata.get("dps")) is None:
            # No DPS metadata — field is assumed always supported
            return True
        return int(dps) in self._product.supported_schema_ids

    async def refresh(self) -> None:
        """Refresh the contents of this trait.

        This will use cached device features if available since they do not
        change often and this avoids unnecessary RPC calls. This would only
        ever change with a firmware update, so caching is appropriate.
        """
        cache_data = await self._device_cache.get()
        if cache_data.device_features is not None:
            common.merge_trait_values(self, cache_data.device_features)
            return
        # Save cached device features
        await super().refresh()
        cache_data.device_features = self
        await self._device_cache.set(cache_data)



"""Tests for code mappings.
These tests exercise the custom enum methods using arbitrary enum values.
"""

import logging

import pytest

from roborock import HomeDataProduct, RoborockCategory
from roborock.data.b01_q10.b01_q10_code_mappings import B01_Q10_DP, YXCleanType
from roborock.data.code_mappings import completed_warnings


def test_from_code() -> None:
    """Test from_code method."""
    assert B01_Q10_DP.START_CLEAN == B01_Q10_DP.from_code(201)
    assert B01_Q10_DP.PAUSE == B01_Q10_DP.from_code(204)
    assert B01_Q10_DP.STOP == B01_Q10_DP.from_code(206)


def test_invalid_from_code() -> None:
    """Test invalid from_code method."""
    with pytest.raises(ValueError, match="999999 is not a valid code for B01_Q10_DP"):
        B01_Q10_DP.from_code(999999)


def test_invalid_from_code_optional() -> None:
    """Test invalid from_code_optional method."""
    assert B01_Q10_DP.from_code_optional(999999) is None


def test_from_code_optional_does_not_warn(caplog: pytest.LogCaptureFixture) -> None:
    """from_code_optional must silently return None for unknown codes.

    Regression test: ss07 hardware pushes data points this library does not model
    (e.g. DPs 112 and 113); resolving them via the optional lookup must not emit
    the "not a valid code" warning that the strict ``from_code`` logs.
    """
    completed_warnings.discard("112 is not a valid code for B01_Q10_DP")
    completed_warnings.discard("113 is not a valid code for B01_Q10_DP")
    with caplog.at_level(logging.WARNING):
        assert B01_Q10_DP.from_code_optional(112) is None
        assert B01_Q10_DP.from_code_optional(113) is None
    assert "not a valid code" not in caplog.text


def test_from_code_still_warns(caplog: pytest.LogCaptureFixture) -> None:
    """The strict from_code keeps logging and raising on unknown codes."""
    completed_warnings.discard("87654 is not a valid code for B01_Q10_DP")
    with caplog.at_level(logging.WARNING), pytest.raises(ValueError, match="not a valid code"):
        B01_Q10_DP.from_code(87654)
    assert "87654 is not a valid code for B01_Q10_DP" in caplog.text


def test_from_name() -> None:
    """Test from_name method."""
    assert B01_Q10_DP.START_CLEAN == B01_Q10_DP.from_name("START_CLEAN")
    assert B01_Q10_DP.PAUSE == B01_Q10_DP.from_name("pause")
    assert B01_Q10_DP.STOP == B01_Q10_DP.from_name("Stop")


def test_invalid_from_name() -> None:
    """Test invalid from_name method."""
    with pytest.raises(ValueError, match="INVALID_NAME is not a valid name for B01_Q10_DP"):
        B01_Q10_DP.from_name("INVALID_NAME")


def test_from_value() -> None:
    """Test from_value method."""
    assert B01_Q10_DP.START_CLEAN == B01_Q10_DP.from_value("dpStartClean")
    assert B01_Q10_DP.PAUSE == B01_Q10_DP.from_value("dpPause")
    assert B01_Q10_DP.STOP == B01_Q10_DP.from_value("dpStop")


def test_invalid_from_value() -> None:
    """Test invalid from_value method."""
    with pytest.raises(ValueError, match="invalid_value is not a valid value for B01_Q10_DP"):
        B01_Q10_DP.from_value("invalid_value")


@pytest.mark.parametrize(
    "input, expected",
    [
        ("START_CLEAN", B01_Q10_DP.START_CLEAN),
        ("start_clean", B01_Q10_DP.START_CLEAN),
        ("dpStartClean", B01_Q10_DP.START_CLEAN),
        (201, B01_Q10_DP.START_CLEAN),
        ("PAUSE", B01_Q10_DP.PAUSE),
        ("pause", B01_Q10_DP.PAUSE),
        ("dpPause", B01_Q10_DP.PAUSE),
        (204, B01_Q10_DP.PAUSE),
        ("STOP", B01_Q10_DP.STOP),
        ("stop", B01_Q10_DP.STOP),
        ("dpStop", B01_Q10_DP.STOP),
        (206, B01_Q10_DP.STOP),
        ("invalid_value", None),
        (999999, None),
    ],
)
def test_from_any_optional(input: str | int, expected: B01_Q10_DP | None) -> None:
    """Test from_any_optional method."""
    assert B01_Q10_DP.from_any_optional(input) == expected


def test_homedata_product_unknown_category():
    """Test that HomeDataProduct can handle unknown categories."""
    data = {
        "id": "unknown_cat_id",
        "name": "Unknown Device",
        "model": "roborock.vacuum.a87",
        "category": "roborock.random.category",
        "schema": [],
    }

    product = HomeDataProduct.from_dict(data)
    assert product.id == "unknown_cat_id"
    assert product.category == RoborockCategory.UNKNOWN


@pytest.mark.parametrize(
    ("readable_value", "expected_clean_type"),
    [
        ("vac_and_mop", YXCleanType.VAC_AND_MOP),
        ("vacuum", YXCleanType.VACUUM),
        ("mop", YXCleanType.MOP),
        ("customized", YXCleanType.CUSTOMIZED),
    ],
)
def test_yx_clean_type_from_value_readable_values(readable_value: str, expected_clean_type: YXCleanType) -> None:
    """Test YXCleanType accepts canonical readable values."""
    assert YXCleanType.from_value(readable_value) is expected_clean_type
    assert expected_clean_type.value == readable_value


def test_yx_clean_type_from_code_customized() -> None:
    """Test YXCleanType accepts custom mode code used by Q10 status updates."""
    assert YXCleanType.from_code(4) is YXCleanType.CUSTOMIZED



"""Tests for the Device class."""

import pathlib
from collections.abc import Callable
from typing import Any
from unittest.mock import AsyncMock, Mock

import pytest
from syrupy import SnapshotAssertion

from roborock.data import HomeData, NetworkInfo, S7MaxVStatus, UserData
from roborock.devices.cache import DeviceCache, DeviceCacheData, InMemoryCache, NoCache
from roborock.devices.device import RoborockDevice
from roborock.devices.rpc.v1_channel import V1Channel
from roborock.devices.traits import v1
from roborock.devices.traits.v1.common import V1TraitMixin
from roborock.devices.transport.local_channel import LocalSession
from roborock.exceptions import RoborockException
from roborock.protocols.v1_protocol import SecurityData, decode_rpc_response
from roborock.roborock_message import RoborockMessage, RoborockMessageProtocol
from tests import mock_data
from tests.fixtures.channel_fixtures import FakeChannel

USER_DATA = UserData.from_dict(mock_data.USER_DATA)
HOME_DATA = HomeData.from_dict(mock_data.HOME_DATA_RAW)
STATUS = S7MaxVStatus.from_dict(mock_data.STATUS)

TESTDATA = pathlib.Path("tests/protocols/testdata/v1_protocol/")


@pytest.fixture(autouse=True, name="channel")
def device_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="rpc_channel")
def rpc_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="mqtt_rpc_channel")
def mqtt_rpc_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="map_rpc_channel")
def map_rpc_channel_fixture() -> AsyncMock:
    """Fixture to set up the channel for tests."""
    return AsyncMock()


@pytest.fixture(autouse=True, name="device")
def device_fixture(channel: AsyncMock, rpc_channel: AsyncMock, mqtt_rpc_channel: AsyncMock) -> RoborockDevice:
    """Fixture to set up the device for tests."""
    return RoborockDevice(
        device_info=HOME_DATA.devices[0],
        product=HOME_DATA.products[0],
        channel=channel,
        trait=v1.create(
            HOME_DATA.devices[0].duid,
            HOME_DATA.products[0],
            HOME_DATA,
            rpc_channel,
            mqtt_rpc_channel,
            AsyncMock(),
            Mock(),
            AsyncMock(),
            device_cache=DeviceCache(HOME_DATA.devices[0].duid, NoCache()),
            region=USER_DATA.region,
        ),
    )


async def test_device_connection(device: RoborockDevice, channel: AsyncMock, setup_rpc_channel: AsyncMock) -> None:
    """Test the Device connection setup."""

    unsub = Mock()
    subscribe = AsyncMock()
    subscribe.return_value = unsub
    channel.subscribe = subscribe

    assert device.duid == "abc123"
    assert device.name == "Roborock S7 MaxV"

    assert not subscribe.called

    await device.connect()
    assert subscribe.called
    assert not unsub.called

    await device.close()
    assert unsub.called


@pytest.mark.parametrize(
    ("connected", "local_connected"),
    [
        (True, False),
        (False, False),
        (True, True),
        (False, True),
    ],
)
async def test_connection_status(
    device: RoborockDevice,
    channel: AsyncMock,
    connected: bool,
    local_connected: bool,
) -> None:
    """Test successful RPC command sending and response handling."""
    channel.is_connected = connected
    channel.is_local_connected = local_connected
    assert device.is_connected is connected
    assert device.is_local_connected is local_connected


@pytest.fixture(name="payloads")
def payloads_fixture() -> list[pathlib.Path]:
    """Fixture to provide the payload for the tests."""
    return []


@pytest.fixture(name="setup_rpc_channel")
def setup_rpc_channel_fixture(rpc_channel: AsyncMock, payloads: list[pathlib.Path]) -> AsyncMock:
    """Fixture to set up the RPC channel for the tests."""
    # Device discovery calls
    side_effects: list[dict[str, Any] | list[Any] | int] = [
        [mock_data.APP_GET_INIT_STATUS],
        mock_data.STATUS,
    ]

    # Subsequent calls return the data payloads setup by the test.
    for payload in payloads:
        # The values other than the payload are arbitrary
        message = RoborockMessage(
            protocol=RoborockMessageProtocol.GENERAL_RESPONSE,
            payload=payload.read_bytes(),
            seq=12750,
            version=b"1.0",
            random=97431,
            timestamp=1652547161,
        )
        response_message = decode_rpc_response(message)
        side_effects.append(response_message.data)

    rpc_channel.send_command.side_effect = side_effects
    return rpc_channel


@pytest.mark.parametrize(
    ("payloads", "property_method"),
    [
        ([TESTDATA / "get_status.json"], lambda x: x.status),
        ([TESTDATA / "get_dnd.json"], lambda x: x.dnd),
        ([TESTDATA / "get_clean_summary.json", TESTDATA / "get_last_clean_record.json"], lambda x: x.clean_summary),
        ([TESTDATA / "get_volume.json"], lambda x: x.sound_volume),
    ],
    ids=[
        "status",
        "dnd",
        "clean_summary",
        "volume",
    ],
)
async def test_device_trait_command_parsing(
    device: RoborockDevice,
    setup_rpc_channel: AsyncMock,
    snapshot: SnapshotAssertion,
    property_method: Callable[..., V1TraitMixin],
) -> None:
    """Test the device trait command."""
    await device.connect()

    trait = property_method(device.v1_properties)
    assert trait
    assert isinstance(trait, V1TraitMixin)
    await trait.refresh()
    assert setup_rpc_channel.send_command.called

    assert trait == snapshot

    assert device.v1_properties
    device_dict = device.diagnostic_data()
    assert device_dict == snapshot


@pytest.mark.parametrize(
    "start_error",
    [RoborockException("transient status fetch failed"), ValueError("unexpected")],
    ids=["roborock-exception", "non-roborock-exception"],
)
async def test_connect_unsubscribes_when_start_fails(
    device: RoborockDevice,
    channel: AsyncMock,
    start_error: Exception,
) -> None:
    """connect() must release the channel when start() fails, for any exception.

    Regression: the cleanup was scoped to ``except RoborockException``, so a
    non-Roborock failure in start() propagated without unsubscribing, leaving the
    channel subscribed and the next attempt unable to re-subscribe.
    """
    unsub = Mock()
    channel.subscribe = AsyncMock(return_value=unsub)
    device.v1_properties.start = AsyncMock(side_effect=start_error)  # type: ignore[method-assign, union-attr]

    with pytest.raises(type(start_error)):
        await device.connect()

    # The channel was released before the error propagated.
    channel.subscribe.assert_awaited_once()
    unsub.assert_called_once()

    # The device is left re-connectable: once start() recovers, connect()
    # succeeds instead of raising "Already connected to the device".
    device.v1_properties.start = AsyncMock(return_value=None)  # type: ignore[method-assign, union-attr]
    await device.connect()
    assert channel.subscribe.await_count == 2


async def test_connect_retries_after_transient_start_failure() -> None:
    """End-to-end regression for the Q5 multi-vacuum bug.

    A device backed by a real V1Channel: the first connect() subscribes, then
    start() fails transiently. The retry must re-subscribe cleanly rather than
    raising "Only one subscription allowed at a time", and the device must end
    up connected.
    """
    duid = HOME_DATA.devices[0].duid

    mqtt_channel = FakeChannel()
    await mqtt_channel.connect()
    local_channel = FakeChannel()
    local_session = Mock(spec=LocalSession, return_value=local_channel)

    # Cache the network info so local connect doesn't need an MQTT round-trip.
    cache = InMemoryCache()
    device_cache = DeviceCache(duid, cache)
    await device_cache.set(DeviceCacheData(network_info=NetworkInfo.from_dict(mock_data.NETWORK_INFO)))

    v1_channel = V1Channel(
        device_uid=duid,
        security_data=SecurityData(endpoint="test_endpoint", nonce=b"test_nonce_16byt"),
        mqtt_channel=mqtt_channel,  # type: ignore[arg-type]
        local_session=local_session,
        device_cache=device_cache,
    )
    device = RoborockDevice(
        device_info=HOME_DATA.devices[0],
        product=HOME_DATA.products[0],
        channel=v1_channel,
        trait=v1.create(
            duid,
            HOME_DATA.products[0],
            HOME_DATA,
            AsyncMock(),
            AsyncMock(),
            AsyncMock(),
            Mock(),
            AsyncMock(),
            device_cache=device_cache,
            region=USER_DATA.region,
        ),
    )

    # First connect() subscribes successfully, then start() fails transiently;
    # the second succeeds.
    device.v1_properties.start = AsyncMock(side_effect=[RoborockException("transient"), None])  # type: ignore[method-assign, union-attr]

    with pytest.raises(RoborockException):
        await device.connect()

    # The retry must NOT raise "Only one subscription allowed at a time"; the
    # clean release after the transient failure lets connect() re-subscribe and
    # the device ends up connected.
    await device.connect()
    assert device.is_connected

    await device.close()



"""Tests for the DoNotDisturbTrait class."""

from unittest.mock import AsyncMock, call

import pytest

from roborock.devices.device import RoborockDevice
from roborock.devices.traits.v1.consumeable import ConsumableAttribute, ConsumableTrait
from roborock.roborock_typing import RoborockCommand

CONSUMABLE_DATA = [
    {
        "main_brush_work_time": 879348,
        "side_brush_work_time": 707618,
        "filter_work_time": 738722,
        "filter_element_work_time": 0,
        "sensor_dirty_time": 455517,
    }
]


@pytest.fixture
def consumable_trait(device: RoborockDevice) -> ConsumableTrait:
    """Create a ConsumableTrait instance with mocked dependencies."""
    assert device.v1_properties
    return device.v1_properties.consumables


async def test_get_consumable_data_success(consumable_trait: ConsumableTrait, mock_rpc_channel: AsyncMock) -> None:
    """Test successfully getting consumable data."""
    # Setup mock to return the sample consumable data
    mock_rpc_channel.send_command.return_value = CONSUMABLE_DATA

    # Call the method
    await consumable_trait.refresh()
    # Verify the result
    assert consumable_trait.main_brush_work_time == 879348
    assert consumable_trait.side_brush_work_time == 707618
    assert consumable_trait.filter_work_time == 738722
    assert consumable_trait.filter_element_work_time == 0
    assert consumable_trait.sensor_dirty_time == 455517

    # Verify the RPC call was made correctly
    mock_rpc_channel.send_command.assert_called_once_with(RoborockCommand.GET_CONSUMABLE)


@pytest.mark.parametrize(
    ("consumable", "reset_param"),
    [
        (ConsumableAttribute.MAIN_BRUSH_WORK_TIME, "main_brush_work_time"),
        (ConsumableAttribute.SIDE_BRUSH_WORK_TIME, "side_brush_work_time"),
        (ConsumableAttribute.FILTER_WORK_TIME, "filter_work_time"),
        (ConsumableAttribute.SENSOR_DIRTY_TIME, "sensor_dirty_time"),
        (ConsumableAttribute.STRAINER_WORK_TIME, "strainer_work_time"),
        (ConsumableAttribute.CLEANING_BRUSH_WORK_TIME, "cleaning_brush_work_time"),
    ],
)
async def test_reset_consumable_data(
    consumable_trait: ConsumableTrait,
    mock_rpc_channel: AsyncMock,
    consumable: ConsumableAttribute,
    reset_param: str,
) -> None:
    """Test successfully resetting consumable data."""
    mock_rpc_channel.send_command.side_effect = [
        {},  # Response for RESET_CONSUMABLE
        # Response for GET_CONSUMABLE after reset
        {
            "main_brush_work_time": 5555,
            "side_brush_work_time": 6666,
            "filter_work_time": 7777,
            "filter_element_work_time": 8888,
            "sensor_dirty_time": 9999,
        },
    ]

    # Call the method
    await consumable_trait.reset_consumable(consumable)

    # Verify the RPC call was made correctly with expected parameters
    assert mock_rpc_channel.send_command.mock_calls == [
        call(RoborockCommand.RESET_CONSUMABLE, params=[reset_param]),
        call(RoborockCommand.GET_CONSUMABLE),
    ]
    # Verify the consumable data was refreshed correctly
    assert consumable_trait.main_brush_work_time == 5555
    assert consumable_trait.side_brush_work_time == 6666
    assert consumable_trait.filter_work_time == 7777
    assert consumable_trait.filter_element_work_time == 8888
    assert consumable_trait.sensor_dirty_time == 9999



"""Common code for MQTT tests."""

import logging
import warnings
from collections.abc import Callable, Generator
from queue import Queue
from typing import Any
from unittest.mock import Mock, patch

import pytest

from .logging import CapturedRequestLog
from .mqtt import FakeMqttSocketHandler

pytest_plugins = [
    "tests.fixtures.logging_fixtures",
]

_LOGGER = logging.getLogger(__name__)

# Used by fixtures to handle incoming requests and prepare responses
MqttRequestHandler = Callable[[bytes], bytes | None]


@pytest.fixture(name="mock_paho_mqtt_create_connection")
def create_connection_fixture(mock_sock: Mock) -> Generator[None, None, None]:
    """Fixture that overrides the MQTT socket creation to wire it up to the mock socket."""
    with patch("paho.mqtt.client.socket.create_connection", return_value=mock_sock):
        yield


@pytest.fixture(name="mock_paho_mqtt_select")
def select_fixture(mock_sock: Mock, fake_mqtt_socket_handler: FakeMqttSocketHandler) -> Generator[None, None, None]:
    """Fixture that overrides the MQTT client select calls to make select work on the mock socket.

    This patch select to activate our mock socket when ready with data. Internal mqtt sockets are
    always ready since they are used internally to wake the select loop. Ours is ready if there
    is data in the buffer.
    """

    def is_ready(sock: Any) -> bool:
        return sock is not mock_sock or (fake_mqtt_socket_handler.pending() > 0)

    def handle_select(rlist: list, wlist: list, *args: Any) -> list:
        return [list(filter(is_ready, rlist)), list(filter(is_ready, wlist))]

    with patch("paho.mqtt.client.select.select", side_effect=handle_select):
        yield


@pytest.fixture(name="fake_mqtt_socket_handler")
def fake_mqtt_socket_handler_fixture(
    mqtt_request_handler: MqttRequestHandler, mqtt_response_queue: Queue[bytes], log: CapturedRequestLog
) -> Generator[FakeMqttSocketHandler, None, None]:
    """Fixture that creates a fake MQTT broker."""
    socket_handler = FakeMqttSocketHandler(mqtt_request_handler, mqtt_response_queue, log)
    yield socket_handler
    if len(socket_handler.response_buf.getvalue()) > 0:
        warnings.warn("Some enqueued MQTT responses were not consumed during the test")


@pytest.fixture(name="mock_sock")
def mock_sock_fixture(fake_mqtt_socket_handler: FakeMqttSocketHandler) -> Mock:
    """Fixture that creates a mock socket connection and wires it to the handler."""
    mock_sock = Mock()
    mock_sock.recv = fake_mqtt_socket_handler.handle_socket_recv
    mock_sock.send = fake_mqtt_socket_handler.handle_socket_send
    mock_sock.pending = fake_mqtt_socket_handler.pending
    return mock_sock


@pytest.fixture(name="mqtt_received_requests")
def received_requests_fixture() -> Queue[bytes]:
    """Fixture that provides access to the received requests."""
    return Queue()


@pytest.fixture(name="mqtt_response_queue")
def response_queue_fixture() -> Generator[Queue[bytes], None, None]:
    """Fixture that provides a queue for enqueueing responses to be sent to the client under test."""
    response_queue: Queue[bytes] = Queue()
    yield response_queue
    if not response_queue.empty():
        warnings.warn("Some enqueued MQTT responses were not consumed during the test")


@pytest.fixture(name="mqtt_request_handler")
def mqtt_request_handler_fixture(
    mqtt_received_requests: Queue[bytes], mqtt_response_queue: Queue[bytes]
) -> MqttRequestHandler:
    """Fixture records incoming requests and replies with responses from the queue."""

    def handle_request(client_request: bytes) -> bytes | None:
        """Handle an incoming request from the client."""
        mqtt_received_requests.put(client_request)

        # Insert a prepared response into the response buffer
        if not mqtt_response_queue.empty():
            return mqtt_response_queue.get()
        return None

    return handle_request



import dataclasses
import datetime
import inspect
import json
import logging
import re
import types
from dataclasses import asdict, dataclass, field
from enum import Enum
from functools import cached_property
from typing import Any, ClassVar, NamedTuple, get_args, get_origin

from .code_mappings import (
    SHORT_MODEL_TO_ENUM,
    RoborockCategory,
    RoborockModeEnum,
    RoborockProductNickname,
)

_LOGGER = logging.getLogger(__name__)


def _camelize(s: str):
    first, *others = s.split("_")
    if len(others) == 0:
        return s
    return "".join([first.lower(), *map(str.title, others)])


def _decamelize(s: str):
    # Split before uppercase letters not at the start, and before numbers
    s = re.sub(r"(?<=[a-z0-9])([A-Z])", r"_\1", s)
    s = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", s)  # Split acronyms followed by normal camelCase
    s = re.sub(r"([a-zA-Z])([0-9]+)", r"\1_\2", s)
    s = s.lower()
    # Temporary fix to avoid breaking any serialization.
    s = s.replace("base_64", "base64")
    return s


def _attr_repr(obj: Any) -> str:
    """Return a string representation of the object including specified attributes.

    This reproduces the default repr behavior of dataclasses, but also includes
    properties. This must be called by the child class's __repr__ method since
    the parent RoborockBase class does not know about the child class's attributes.
    """
    # Reproduce default repr behavior
    parts = []
    for k in dir(obj):
        if k.startswith("_"):
            continue
        try:
            v = getattr(obj, k)
        except (RuntimeError, Exception):
            continue
        if callable(v):
            continue
        parts.append(f"{k}={v!r}")
    return f"{type(obj).__name__}({', '.join(parts)})"


@dataclass(repr=False)
class RoborockBase:
    """Base class for all Roborock data classes."""

    _missing_logged: ClassVar[set[str]] = set()

    @staticmethod
    def _convert_to_class_obj(class_type: type, value):
        if get_origin(class_type) is list:
            sub_type = get_args(class_type)[0]
            return [RoborockBase._convert_to_class_obj(sub_type, obj) for obj in value]
        if get_origin(class_type) is dict:
            key_type, value_type = get_args(class_type)
            if key_type is not None:
                return {key_type(k): RoborockBase._convert_to_class_obj(value_type, v) for k, v in value.items()}
            return {k: RoborockBase._convert_to_class_obj(value_type, v) for k, v in value.items()}
        if inspect.isclass(class_type):
            if issubclass(class_type, RoborockBase):
                return class_type.from_dict(value)
            if issubclass(class_type, RoborockModeEnum):
                return class_type.from_code(value)
        if class_type is Any or type(class_type) is str:
            return value
        return class_type(value)  # type: ignore[call-arg]

    @classmethod
    def from_dict(cls, data: dict[str, Any]):
        """Create an instance of the class from a dictionary."""
        if not isinstance(data, dict):
            return None
        field_types = {field.name: field.type for field in dataclasses.fields(cls)}
        normalized_data: dict[str, Any] = {}
        for orig_key, value in data.items():
            key = _decamelize(orig_key)
            if field_types.get(key) is None:
                if (log_key := f"{cls.__name__}.{key}") not in RoborockBase._missing_logged:
                    _LOGGER.debug(
                        "Key '%s' (decamelized: '%s') not found in %s fields, skipping",
                        orig_key,
                        key,
                        cls.__name__,
                    )
                    RoborockBase._missing_logged.add(log_key)
                continue
            normalized_data[key] = value

        result = RoborockBase.convert_dict(field_types, normalized_data)
        return cls(**result)

    @staticmethod
    def convert_dict(types_map: dict[Any, type], data: dict[Any, Any]) -> dict[Any, Any]:
        """Generic helper to convert a dictionary of values based on a schema map of types.

        This is meant to be used by traits that use dataclass reflection similar to
        `Roborock.from_dict` to merge in new data updates.
        """
        result: dict[Any, Any] = {}
        for key, value in data.items():
            if key not in types_map:
                continue
            field_type = types_map[key]
            if value == "None" or value is None:
                result[key] = None
                continue
            if isinstance(field_type, types.UnionType):
                for subtype in get_args(field_type):
                    if subtype is types.NoneType:
                        continue
                    try:
                        result[key] = RoborockBase._convert_to_class_obj(subtype, value)
                        break
                    except Exception:
                        _LOGGER.exception(f"Failed to convert {key} with value {value} to type {subtype}")
                        continue
            else:
                try:
                    result[key] = RoborockBase._convert_to_class_obj(field_type, value)
                except Exception:
                    _LOGGER.exception(f"Failed to convert {key} with value {value} to type {field_type}")
                    continue

        return result

    def as_dict(self) -> dict:
        return asdict(
            self,
            dict_factory=lambda _fields: {
                _camelize(key): value.value if isinstance(value, Enum) else value
                for (key, value) in _fields
                if value is not None
            },
        )


@dataclass
class RoborockBaseTimer(RoborockBase):
    start_hour: int | None = None
    start_minute: int | None = None
    end_hour: int | None = None
    end_minute: int | None = None
    enabled: int | None = None

    @property
    def start_time(self) -> datetime.time | None:
        return (
            datetime.time(hour=self.start_hour, minute=self.start_minute)
            if self.start_hour is not None and self.start_minute is not None
            else None
        )

    @property
    def end_time(self) -> datetime.time | None:
        return (
            datetime.time(hour=self.end_hour, minute=self.end_minute)
            if self.end_hour is not None and self.end_minute is not None
            else None
        )

    def as_list(self) -> list:
        return [self.start_hour, self.start_minute, self.end_hour, self.end_minute]

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class Reference(RoborockBase):
    r: str | None = None
    a: str | None = None
    m: str | None = None
    l: str | None = None


@dataclass
class RRiot(RoborockBase):
    u: str
    s: str
    h: str
    k: str
    r: Reference


@dataclass
class UserData(RoborockBase):
    rriot: RRiot
    uid: int | None = None
    tokentype: str | None = None
    token: str | None = None
    rruid: str | None = None
    region: str | None = None
    countrycode: str | None = None
    country: str | None = None
    nickname: str | None = None
    tuya_device_state: int | None = None
    avatarurl: str | None = None


@dataclass
class HomeDataProductSchema(RoborockBase):
    id: Any | None = None
    name: Any | None = None
    code: Any | None = None
    mode: Any | None = None
    type: Any | None = None
    product_property: Any | None = None
    property: Any | None = None
    desc: Any | None = None


@dataclass
class HomeDataProduct(RoborockBase):
    id: str
    name: str
    model: str
    category: RoborockCategory
    code: str | None = None
    icon_url: str | None = None
    attribute: Any | None = None
    capability: int | None = None
    schema: list[HomeDataProductSchema] | None = None

    @property
    def product_nickname(self) -> RoborockProductNickname:
        return SHORT_MODEL_TO_ENUM.get(self.model.split(".")[-1], RoborockProductNickname.PEARLPLUS)

    def summary_info(self) -> str:
        """Return a string with key product information for logging purposes."""
        return f"{self.name} (model={self.model}, category={self.category})"

    @cached_property
    def supported_schema_codes(self) -> set[str]:
        """Return a set of schema codes that are supported by the device.

        These correspond with string field names like "state" or "error_code" that
        correspond to RoborockDataProtocol or RoborockB01Protocol code values.
        """
        if self.schema is None:
            return set()
        return {schema.code for schema in self.schema if schema.code is not None}

    @cached_property
    def supported_schema_ids(self) -> set[int]:
        """Return a set of schema IDs (DPS integers) that are supported by the device.

        These correspond to RoborockMessageProtocol and RoborockDataProtocol or
        RoborockB01Protocol enum number values (depends on the device protocol versions).
        """
        if self.schema is None:
            return set()
        return {int(schema.id) for schema in self.schema if schema.id is not None}


@dataclass
class HomeDataDevice(RoborockBase):
    duid: str
    name: str
    local_key: str
    product_id: str
    fv: str | None = None
    attribute: Any | None = None
    active_time: int | None = None
    runtime_env: Any | None = None
    time_zone_id: str | None = None
    icon_url: str | None = None
    lon: Any | None = None
    lat: Any | None = None
    share: Any | None = None
    share_time: Any | None = None
    online: bool | None = None
    pv: str | None = None
    room_id: Any | None = None
    tuya_uuid: Any | None = None
    tuya_migrated: bool | None = None
    extra: Any | None = None
    sn: str | None = None
    feature_set: str | None = None
    new_feature_set: str | None = None
    device_status: dict | None = None
    silent_ota_switch: bool | None = None
    setting: Any | None = None
    f: bool | None = None
    create_time: int | None = None
    cid: str | None = None
    share_type: Any | None = None
    share_expired_time: int | None = None

    def summary_info(self) -> str:
        """Return a string with key device information for logging purposes."""
        return f"{self.name} (pv={self.pv}, fv={self.fv}, online={self.online})"


@dataclass
class HomeDataRoom(RoborockBase):
    id: int
    name: str

    @property
    def iot_id(self) -> str:
        """Return the room's ID as a string IOT ID."""
        return str(self.id)


@dataclass
class HomeDataScene(RoborockBase):
    id: int
    name: str


@dataclass
class HomeDataSchedule(RoborockBase):
    id: int
    cron: str
    repeated: bool
    enabled: bool
    param: dict | None = None


@dataclass
class HomeData(RoborockBase):
    id: int
    name: str
    products: list[HomeDataProduct] = field(default_factory=lambda: [])
    devices: list[HomeDataDevice] = field(default_factory=lambda: [])
    received_devices: list[HomeDataDevice] = field(default_factory=lambda: [])
    lon: Any | None = None
    lat: Any | None = None
    geo_name: Any | None = None
    rooms: list[HomeDataRoom] = field(default_factory=list)

    def get_all_devices(self) -> list[HomeDataDevice]:
        devices = []
        if self.devices is not None:
            devices += self.devices
        if self.received_devices is not None:
            devices += self.received_devices
        return devices

    @cached_property
    def product_map(self) -> dict[str, HomeDataProduct]:
        """Returns a dictionary of product IDs to HomeDataProduct objects."""
        return {product.id: product for product in self.products}

    @cached_property
    def device_products(self) -> dict[str, tuple[HomeDataDevice, HomeDataProduct]]:
        """Returns a dictionary of device DUIDs to HomeDataDeviceProduct objects."""
        product_map = self.product_map
        return {
            device.duid: (device, product)
            for device in self.get_all_devices()
            if (product := product_map.get(device.product_id)) is not None
        }

    @property
    def rooms_map(self) -> dict[str, HomeDataRoom]:
        """Returns a dictionary of Room iot_id to rooms"""
        return {room.iot_id: room for room in self.rooms}

    @property
    def rooms_name_map(self) -> dict[str, str]:
        """Returns a dictionary of Room iot_id to room names."""
        return {room.iot_id: room.name for room in self.rooms}


@dataclass
class LoginData(RoborockBase):
    user_data: UserData
    email: str
    home_data: HomeData | None = None


@dataclass
class DeviceData(RoborockBase):
    device: HomeDataDevice
    model: str
    host: str | None = None

    @property
    def product_nickname(self) -> RoborockProductNickname:
        return SHORT_MODEL_TO_ENUM.get(self.model.split(".")[-1], RoborockProductNickname.PEARLPLUS)

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class RoomMapping(RoborockBase):
    segment_id: int
    iot_id: str


@dataclass
class NamedRoomMapping(RoomMapping):
    """Dataclass representing a mapping of a room segment to a name.

    The name information is not provided by the device directly, but is provided
    from the HomeData based on the iot_id from the room.
    """

    @property
    def name(self) -> str:
        """The human-readable name of the room, or a default name if not available."""
        return self.raw_name or f"Room {self.segment_id}"

    raw_name: str | None = None
    """The raw name of the room, as provided by the device."""


@dataclass
class CombinedMapInfo(RoborockBase):
    """Data structure for caching home information.

    This is not provided directly by the API, but is a combination of map data
    and room data to provide a more useful structure.
    """

    map_flag: int
    """The map identifier."""

    name: str
    """The name of the map from MultiMapsListMapInfo."""

    rooms: list[NamedRoomMapping]
    """The list of rooms in the map."""

    @property
    def rooms_map(self) -> dict[int, NamedRoomMapping]:
        """Returns a mapping of segment_id to NamedRoomMapping."""
        return {room.segment_id: room for room in self.rooms}


@dataclass
class BroadcastMessage(RoborockBase):
    duid: str
    ip: str
    version: bytes


class ServerTimer(NamedTuple):
    id: str
    status: str
    dontknow: int


@dataclass
class RoborockProductStateValue(RoborockBase):
    value: list
    desc: dict


@dataclass
class RoborockProductState(RoborockBase):
    dps: int
    desc: dict
    value: list[RoborockProductStateValue]


@dataclass
class RoborockProductSpec(RoborockBase):
    state: RoborockProductState
    battery: dict | None = None
    dry_countdown: dict | None = None
    extra: dict | None = None
    offpeak: dict | None = None
    countdown: dict | None = None
    mode: dict | None = None
    ota_nfo: dict | None = None
    pause: dict | None = None
    program: dict | None = None
    shutdown: dict | None = None
    washing_left: dict | None = None


@dataclass
class RoborockProduct(RoborockBase):
    id: int | None = None
    name: str | None = None
    model: str | None = None
    packagename: str | None = None
    ssid: str | None = None
    picurl: str | None = None
    cardpicurl: str | None = None
    mediumCardpicurl: str | None = None
    resetwifipicurl: str | None = None
    configPicUrl: str | None = None
    pluginPicUrl: str | None = None
    resetwifitext: dict | None = None
    tuyaid: str | None = None
    status: int | None = None
    rriotid: str | None = None
    pictures: list | None = None
    ncMode: str | None = None
    scope: str | None = None
    product_tags: list | None = None
    agreements: list | None = None
    cardspec: str | None = None
    plugin_pic_url: str | None = None

    @property
    def product_nickname(self) -> RoborockProductNickname | None:
        if self.cardspec:
            return RoborockProductSpec.from_dict(json.loads(self.cardspec).get("data"))
        return None

    def __repr__(self) -> str:
        return _attr_repr(self)


@dataclass
class RoborockProductCategory(RoborockBase):
    id: int
    display_name: str
    icon_url: str


@dataclass
class RoborockCategoryDetail(RoborockBase):
    category: RoborockProductCategory
    product_list: list[RoborockProduct]


@dataclass
class ProductResponse(RoborockBase):
    category_detail_list: list[RoborockCategoryDetail]



from roborock.data.b01_q7.b01_q7_containers import Q7MapListEntry
from roborock.devices.traits.b01.q7 import Q7PropertiesApi
from roborock.roborock_typing import RoborockB01Q7Methods

from .conftest import FakeQ7Channel


async def test_q7_map_refresh(q7_api: Q7PropertiesApi, fake_channel: FakeQ7Channel):
    """Test retrieving lists of saved maps."""
    fake_channel.response_queue.append({"map_list": [{"id": 111, "name": "Map 1"}]})
    await q7_api.map.refresh()

    assert len(fake_channel.published_commands) == 1
    command, params = fake_channel.published_commands[0]
    assert command == RoborockB01Q7Methods.GET_MAP_LIST
    assert params == {}
    assert q7_api.map.map_list == [Q7MapListEntry(id=111)]



"""Map trait for B01 Q7 devices."""

from roborock.data import Q7MapList
from roborock.devices.rpc.b01_q7_channel import Q7RpcChannel
from roborock.devices.traits import Trait
from roborock.exceptions import RoborockException
from roborock.roborock_typing import RoborockB01Q7Methods


class MapTrait(Q7MapList, Trait):
    """Map trait for B01/Q7 devices, responsible for fetching and caching map list metadata.

    The MapContent is fetched from the MapContent trait, which relies on this trait to determine the
    current map ID to fetch.
    """

    def __init__(self, channel: Q7RpcChannel) -> None:
        super().__init__()
        self._channel = channel

    async def refresh(self) -> None:
        """Refresh cached map list metadata from the device."""
        response = await self._channel.send_command(
            command=RoborockB01Q7Methods.GET_MAP_LIST,
            params={},
        )
        if not isinstance(response, dict):
            raise RoborockException(
                f"Unexpected response type for GET_MAP_LIST: {type(response).__name__}: {response!r}"
            )

        if (parsed := Q7MapList.from_dict(response)) is None:
            raise RoborockException(f"Failed to decode map list response: {response!r}")

        self.map_list = parsed.map_list



import datetime
import json
from dataclasses import dataclass, field
from functools import cached_property

from ...exceptions import RoborockException
from ..containers import RoborockBase
from .b01_q7_code_mappings import (
    B01Fault,
    CleanPathPreferenceMapping,
    CleanRepeatMapping,
    CleanTypeMapping,
    SCWindMapping,
    WaterLevelMapping,
    WorkModeMapping,
    WorkStatusMapping,
)


@dataclass
class NetStatus(RoborockBase):
    """Represents the network status of the device."""

    rssi: str
    loss: int
    ping: int
    ip: str
    mac: str
    ssid: str
    frequency: int
    bssid: str


@dataclass
class OrderTotal(RoborockBase):
    """Represents the order total information."""

    total: int
    enable: int


@dataclass
class Privacy(RoborockBase):
    """Represents the privacy settings of the device."""

    ai_recognize: int
    dirt_recognize: int
    pet_recognize: int
    carpet_turbo: int
    carpet_avoid: int
    carpet_show: int
    map_uploads: int
    ai_agent: int
    ai_avoidance: int
    record_uploads: int
    along_floor: int
    auto_upgrade: int


@dataclass
class PvCharging(RoborockBase):
    """Represents the photovoltaic charging status."""

    status: int
    begin_time: int
    end_time: int


@dataclass
class Recommend(RoborockBase):
    """Represents cleaning recommendations."""

    sill: int
    wall: int
    room_id: list[int] = field(default_factory=list)


@dataclass
class Q7MapListEntry(RoborockBase):
    """Single map list entry returned by `service.get_map_list`."""

    id: int | None = None
    cur: bool | None = None


@dataclass
class Q7MapList(RoborockBase):
    """Map list response returned by `service.get_map_list`."""

    map_list: list[Q7MapListEntry] = field(default_factory=list)

    @property
    def current_map_id(self) -> int | None:
        """Current map id, preferring the entry marked current."""
        if not self.map_list:
            return None

        ordered = sorted(self.map_list, key=lambda entry: entry.cur or False, reverse=True)
        first = next(iter(ordered), None)
        if first is None or not isinstance(first.id, int):
            return None
        return first.id


@dataclass
class B01Props(RoborockBase):
    """
    Represents the complete properties and status for a Roborock B01 model.
    This dataclass is generated based on the device's status JSON object.
    """

    status: WorkStatusMapping | None = None
    fault: B01Fault | None = None
    wind: SCWindMapping | None = None
    water: WaterLevelMapping | None = None
    mode: CleanTypeMapping | None = None
    quantity: int | None = None  # The Q7 L5 reports its battery level as 'quantity'
    alarm: int | None = None
    volume: int | None = None
    hypa: int | None = None
    main_brush: int | None = None
    side_brush: int | None = None
    mop_life: int | None = None
    main_sensor: int | None = None
    net_status: NetStatus | None = None
    repeat_state: CleanRepeatMapping | None = None
    tank_state: int | None = None
    sweep_type: int | None = None
    clean_path_preference: CleanPathPreferenceMapping | None = None
    cloth_state: int | None = None
    time_zone: int | None = None
    time_zone_info: str | None = None
    language: int | None = None
    cleaning_time: int | None = None
    real_clean_time: int | None = None
    cleaning_area: int | None = None
    custom_type: int | None = None
    sound: int | None = None
    work_mode: WorkModeMapping | None = None
    station_act: int | None = None
    charge_state: int | None = None
    current_map_id: int | None = None
    map_num: int | None = None
    dust_action: int | None = None
    quiet_is_open: int | None = None
    quiet_begin_time: int | None = None
    quiet_end_time: int | None = None
    clean_finish: int | None = None
    voice_type: int | None = None
    voice_type_version: int | None = None
    order_total: OrderTotal | None = None
    build_map: int | None = None
    privacy: Privacy | None = None
    dust_auto_state: int | None = None
    dust_frequency: int | None = None
    child_lock: int | None = None
    multi_floor: int | None = None
    map_save: int | None = None
    light_mode: int | None = None
    green_laser: int | None = None
    dust_bag_used: int | None = None
    order_save_mode: int | None = None
    manufacturer: str | None = None
    back_to_wash: int | None = None
    charge_station_type: int | None = None
    pv_cut_charge: int | None = None
    pv_charging: PvCharging | None = None
    serial_number: str | None = None
    recommend: Recommend | None = None
    add_sweep_status: int | None = None

    @property
    def battery(self) -> int | None:
        """
        Returns device battery level as a percentage.
        """
        return self.quantity

    @property
    def main_brush_time_left(self) -> int | None:
        """
        Returns estimated remaining life of the main brush in minutes.
        Total life is 300 hours (18000 minutes).
        """
        if self.main_brush is None:
            return None
        return max(0, 18000 - self.main_brush)

    @property
    def side_brush_time_left(self) -> int | None:
        """
        Returns estimated remaining life of the side brush in minutes.
        Total life is 200 hours (12000 minutes).
        """
        if self.side_brush is None:
            return None
        return max(0, 12000 - self.side_brush)

    @property
    def filter_time_left(self) -> int | None:
        """
        Returns estimated remaining life of the filter (hypa) in minutes.
        Total life is 150 hours (9000 minutes).
        """
        if self.hypa is None:
            return None
        return max(0, 9000 - self.hypa)

    @property
    def mop_life_time_left(self) -> int | None:
        """
        Returns estimated remaining life of the mop in minutes.
        Total life is 180 hours (10800 minutes).
        """
        if self.mop_life is None:
            return None
        return max(0, 10800 - self.mop_life)

    @property
    def sensor_dirty_time_left(self) -> int | None:
        """
        Returns estimated time until sensors need cleaning in minutes.
        Maintenance interval is typically 30 hours (1800 minutes).
        """
        if self.main_sensor is None:
            return None
        return max(0, 1800 - self.main_sensor)

    @property
    def status_name(self) -> str | None:
        """Returns the name of the current status."""
        return self.status.value if self.status is not None else None

    @property
    def fault_name(self) -> str | None:
        """Returns the name of the current fault."""
        return self.fault.value if self.fault is not None else None

    @property
    def wind_name(self) -> str | None:
        """Returns the name of the current fan speed (wind)."""
        return self.wind.value if self.wind is not None else None

    @property
    def work_mode_name(self) -> str | None:
        """Returns the name of the current work mode."""
        return self.work_mode.value if self.work_mode is not None else None

    @property
    def repeat_state_name(self) -> str | None:
        """Returns the name of the current repeat state."""
        return self.repeat_state.value if self.repeat_state is not None else None

    @property
    def clean_path_preference_name(self) -> str | None:
        """Returns the name of the current clean path preference."""
        return self.clean_path_preference.value if self.clean_path_preference is not None else None


@dataclass
class CleanRecordDetail(RoborockBase):
    """Represents a single clean record detail (from `record_list[].detail`)."""

    record_start_time: int | None = None
    method: int | None = None
    record_use_time: int | None = None
    clean_count: int | None = None
    # This is seemingly returned in meters (non-squared)
    record_clean_area: int | None = None
    record_clean_mode: int | None = None
    record_clean_way: int | None = None
    record_task_status: int | None = None
    record_faultcode: int | None = None
    record_dust_num: int | None = None
    clean_current_map: int | None = None
    record_map_url: str | None = None

    @property
    def start_datetime(self) -> datetime.datetime | None:
        """Convert the start datetime into a datetime object."""
        if self.record_start_time is not None:
            return datetime.datetime.fromtimestamp(self.record_start_time).astimezone(datetime.UTC)
        return None

    @property
    def square_meters_area_cleaned(self) -> float | None:
        """Returns the area cleaned in square meters."""
        if self.record_clean_area is not None:
            return self.record_clean_area / 100
        return None


@dataclass
class CleanRecordListItem(RoborockBase):
    """Represents an entry in the clean record list returned by `service.get_record_list`."""

    url: str | None = None
    detail: str | None = None

    @cached_property
    def detail_parsed(self) -> CleanRecordDetail | None:
        """Parse and return the detail as a CleanRecordDetail object."""
        if self.detail is None:
            return None
        try:
            parsed = json.loads(self.detail)
        except json.JSONDecodeError as ex:
            raise RoborockException(f"Invalid B01 record detail JSON: {self.detail!r}") from ex
        return CleanRecordDetail.from_dict(parsed)


@dataclass
class CleanRecordList(RoborockBase):
    """Represents the clean record list response from `service.get_record_list`."""

    total_area: int | None = None
    total_time: int | None = None  # stored in seconds
    total_count: int | None = None
    record_list: list[CleanRecordListItem] = field(default_factory=list)

    @property
    def square_meters_area_cleaned(self) -> float | None:
        """Returns the area cleaned in square meters."""
        if self.total_area is not None:
            return self.total_area / 100
        return None


@dataclass
class CleanRecordSummary(RoborockBase):
    """Represents clean record totals for B01/Q7 devices."""

    total_time: int | None = None
    total_area: int | None = None
    total_count: int | None = None
    last_record_detail: CleanRecordDetail | None = None
